# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import os
from collections import namedtuple

import simplejson
from Crypto.Cipher import AES

from data_pipeline.config import get_config
from data_pipeline.helpers.decorators import memoized
from data_pipeline.helpers.singleton import Singleton
from data_pipeline.initialization_vector import get_initialization_vector
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer


_AVSCInfo = namedtuple('_AVSCInfo', (
    'id',
    'avsc_file_path',
    'namespace',
    'source',
    'source_owner_email',
    'contains_pii'
))


initialization_vector_info = _AVSCInfo(
    1,
    'data_pipeline/schemas/initialization_vector_v1.avsc',
    'yelp.data_pipeline',
    'initialization_vector',
    'bam+data_pipeline@yelp.com',
    False
)


class _AVSCStore(object):
    """Services/Applications are responsible for registering the
    meta attribute avro schemas and caching them if necessary.
    For the meta attributes such as encryption
    (or initialization_vector) which is added by the clientlib
    internally, this class is then designed to register
    and cache the avro schemas for such meta attributes inside
    the clientlib.

    This may be replcaced with something better in the future. This class
    is not meant to be used outside of this file.
    """

    __metaclass__ = Singleton

    def __init__(self):
        self._schematizer = get_schematizer()
        self._schema_id_cache = {}

    def get_schema_id(self, avro_schema_info):
        key = avro_schema_info.id
        schema_id = self._schema_id_cache.get(key)
        if not schema_id:
            schema_id = self._load_schema(avro_schema_info)
        return schema_id

    def update_schema_cache(self, avro_schema_info, schema_id):
        self._schema_id_cache[avro_schema_info.id] = schema_id

    def _load_schema(self, avro_schema_info):
        schema_info = self._register_schema(avro_schema_info)
        self.update_schema_cache(avro_schema_info, schema_info.schema_id)
        return schema_info.schema_id

    def _load_avro_schema_file(self, avsc_file_path):
        schema_path = os.path.join(
            os.path.dirname(__file__),
            os.pardir,
            avsc_file_path
        )
        with open(schema_path, 'r') as f:
            return simplejson.loads(f.read())

    def _register_schema(self, avro_schema_info):
        avro_schema_json = self._load_avro_schema_file(
            avro_schema_info.avsc_file_path
        )
        return self._schematizer.register_schema_from_schema_json(
            namespace=avro_schema_info.namespace,
            source=avro_schema_info.source,
            schema_json=avro_schema_json,
            source_owner_email=avro_schema_info.source_owner_email,
            contains_pii=avro_schema_info.contains_pii
        )


class EncryptionHelper(object):
    """The EncryptionHelper provides helper methods for encrypting message
    payload in the data pipeline.

    Args:
        encryption_type (string): string indicating the encryption algorithm and
            encryption key used to encrypt/decrypt payload.  The format must be
            "{algorithm_name}-{key_id}".  The "key_id" is the file name that
            contains specific encryption key for the specified algorithm.
        encryption_meta (:class:data_pipeline.meta_attribute.MetaAttribute):
            meta attribute that contains necessary information for the given
            encryption type to perform encryption/decryption.  For example, if
            the encryption type is AES algorithm, the meta attribute is then
            the initialization vector meta attribute.

    Remarks:
        This class currently is implemented specifically for AES algorithm,
        although the original design is to support multiple encryption algorithms.
    """

    def __init__(self, encryption_type, encryption_meta=None):
        key_location = get_config().key_location + 'key-{}.key'
        self.key = self._retrieve_key(encryption_type, key_location)
        self.encryption_meta = (
            encryption_meta or
            self.get_encryption_meta_by_encryption_type(encryption_type)
        )

    def _retrieve_key(self, encryption_type, key_location):
        if not encryption_type:
            raise ValueError("Encryption type should be set.")

        # Get the key number to use, allowing for key rotation.
        _, key_id = self._get_algorithm_and_key_id(encryption_type)

        return fetch_encyption_key(key_location.format(key_id))

    @classmethod
    def _get_algorithm_and_key_id(cls, encryption_type):
        # encryption_type must be of the form 'Algorithm_name-{key_id}'
        algorithm, key_id = encryption_type.split('-')
        return algorithm, key_id

    @classmethod
    def get_encryption_meta_by_encryption_type(cls, encryption_type):
        """This function returns the meta attribute for the given encryption type.

        Remarks:
            Currently because only AES algorithm is supported, it returns
            `class:data_pipeline.initialization_vector.InitializationVector` meta
            attribute directly.
        """
        algorithm, _ = cls._get_algorithm_and_key_id(encryption_type)
        if algorithm:
            schema_id = _AVSCStore().get_schema_id(initialization_vector_info)
            return get_initialization_vector(schema_id)
        raise Exception(
            "Encryption algorithm {} is not supported.".format(algorithm)
        )

    def encrypt_payload(self, payload):
        """Encrypt payload with key on machine, using AES."""
        encrypter = AES.new(
            self.key,
            AES.MODE_CBC,
            self.encryption_meta.payload
        )
        payload = self._pad_payload(payload)
        return encrypter.encrypt(payload)

    def _pad_payload(self, payload):
        """payloads must have length equal to a multiple of 16 in order to
        be encrypted by AES's CBC algorithm, because it uses block chaining.
        This method adds a chr equal to the length needed in bytes, bytes times,
        to the end of payload before encrypting it, and the _unpad method
        removes those bytes.
        """
        length = 16 - (len(payload) % 16)
        return payload + chr(length) * length

    def decrypt_payload(self, payload):
        decrypter = AES.new(
            self.key,
            AES.MODE_CBC,
            self.encryption_meta.payload
        )
        data = decrypter.decrypt(payload)
        return self._unpad(data)

    def _unpad(self, payload):
        return payload[:-ord(payload[len(payload) - 1:])]


@memoized
def fetch_encyption_key(file_name):
    with open(file_name, 'r') as f:
        return f.read(AES.block_size)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import base64
import os

import avro.io
import avro.schema
from cached_property import cached_property
from data_pipeline_avro_util.avro_string_reader import AvroStringReader
from data_pipeline_avro_util.avro_string_writer import AvroStringWriter


class Envelope(object):
    """Envelope used to encode and identify a message for transport.

    Envelope instances are meant to be long-lived and used to encode multiple
    messages.

    Example:
        >>> from data_pipeline.message import CreateMessage
        >>> message = CreateMessage(schema_id=1, payload=bytes("FAKE MESSAGE"))
        >>> envelope = Envelope()
        >>> packed_message = envelope.pack(message)
        >>> isinstance(packed_message, bytes)
        True
        >>> unpacked = envelope.unpack(packed_message)
        >>> unpacked['message_type']
        u'create'
        >>> unpacked['schema_id']
        1
        >>> unpacked['payload']
        'FAKE MESSAGE'
    """

    # Magic byte value of packed message specifying that it is base64 encoded.
    # This value was chosen because it is valid ASCII
    ASCII_MAGIC_BYTE = bytes('a')

    @cached_property
    def _schema(self):
        # Keeping this as an instance method because of issues with sharing
        # this data across processes.
        schema_path = os.path.join(
            os.path.dirname(__file__),
            'schemas/envelope_v1.avsc'
        )
        return avro.schema.parse(open(schema_path).read())

    @cached_property
    def _avro_string_writer(self):
        return AvroStringWriter(self._schema)

    @cached_property
    def _avro_string_reader(self):
        return AvroStringReader(self._schema, self._schema)

    def pack(self, message, ascii_encoded=False):
        """Packs a message for transport as described in y/cep342.

        Use :func:`unpack` to decode the packed message.

        Args:
            message (data_pipeline.message.Message): The message to pack
            ascii_decoded (Optional[bool]): Set to True if message is not valid ASCII

        Returns:
            bytes: Avro byte string prepended by magic envelope version byte

        The initial "magic byte" is meant to specify the envelope schema version.
        See y/cep342 for details.  In other words, the version number of the current
        schema is the null byte.  In the event we need to add additional envelope
        versions, we'll use this byte to identify it.

        In addition, the "magic byte" is used as a protocol to encode the serialized
        message in base64. See DATAPIPE-1350 for more detail. This option has been
        added because as of now, yelp_clog only supports sending valid ASCII strings.
        Producer/Consumer registration will make use of this to instead send base64
        encoded strings.
        """
        msg = bytes(0) + self._avro_string_writer.encode(message.avro_repr)

        if ascii_encoded:
            return self.ASCII_MAGIC_BYTE + base64.urlsafe_b64encode(msg)
        else:
            return msg

    def unpack(self, packed_message):
        """Decodes a message packed with :func:`pack`.

        Warning:
            The public API for this function may change to return
            :class:`data_pipeline.message.Message` instances.

        Args:
            packed_message (bytes): The previously packed message

        Returns:
            dict: A dictionary with the decoded Avro representation.
        """

        # If the magic byte is ASCII_MAGIC_BYTE, decode it from base64 to ASCII
        if packed_message[0] == self.ASCII_MAGIC_BYTE:
            packed_message = base64.urlsafe_b64decode(packed_message[1:])

        return self._avro_string_reader.decode(packed_message[1:])



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel


"""
Represent the data of a data target.  A data target represents a destination
where the data(messages) are eventually sent to, such as a Redshift cluster.

Args:
    data_target_id (int): The id of the data target.
    target_type (str): The target type, such as Redshift, etc.
    destination (str): The actual location of the data target, such as the Url
        of a Redshift cluster.
"""
DataTarget = namedtuple(
    'DataTarget',
    ['data_target_id', 'name', 'target_type', 'destination']
)


class _DataTarget(BaseModel):
    """Internal class used to convert from/to various data structure and
    facilitate constructing the return value of schematizer functions.
    """

    def __init__(self, data_target_id, name, target_type, destination):
        self.data_target_id = data_target_id
        self.name = name
        self.target_type = target_type
        self.destination = destination

    @classmethod
    def from_response(cls, response):
        return cls(
            data_target_id=response.data_target_id,
            name=response.name,
            target_type=response.target_type,
            destination=response.destination
        )

    def to_cache_value(self):
        return {
            'data_target_id': self.data_target_id,
            'name': self.name,
            'target_type': self.target_type,
            'destination': self.destination
        }

    @classmethod
    def from_cache_value(cls, cache_value):
        return cls(
            data_target_id=cache_value['data_target_id'],
            name=cache_value['name'],
            target_type=cache_value['target_type'],
            destination=cache_value['destination']
        )

    def to_result(self):
        return DataTarget(
            data_target_id=self.data_target_id,
            name=self.name,
            target_type=self.target_type,
            destination=self.destination
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel
from data_pipeline.schematizer_clientlib.models.namespace import _Namespace


"""
Represent the data of a source.  Source is a sub-group under namespaces which
an avro schema is created for.  For example, `user` (table) could be a source.

Args:
    source_id (int): The id of the source.
    name (str): The name of the source.
    owner_email (str): The email of the source owner.
    namespace (data_pipeline.schematizer_clientlib.models.namespace.Namespace):
        The namespace of the source.
    category (str): The category of the source. (e.g. Content, Deals etc.)
"""
Source = namedtuple(
    'Source',
    ['source_id', 'name', 'owner_email', 'namespace', 'category']
)


class _Source(BaseModel):

    def __init__(self, source_id, name, owner_email, namespace, category):
        self.source_id = source_id
        self.name = name
        self.owner_email = owner_email
        self.namespace = namespace
        self.category = category

    @classmethod
    def from_response(cls, response):
        return cls(
            source_id=response.source_id,
            name=response.name,
            owner_email=response.owner_email,
            namespace=_Namespace.from_response(response.namespace),
            category=response.category
        )

    def to_cache_value(self):
        return {
            'source_id': self.source_id,
            'name': self.name,
            'owner_email': self.owner_email,
            'namespace': self.namespace,
            'category': self.category
        }

    @classmethod
    def from_cache_value(cls, cache_value):
        return cls(
            source_id=cache_value['source_id'],
            name=cache_value['name'],
            owner_email=cache_value['owner_email'],
            namespace=cache_value['namespace'],
            category=cache_value['category']
        )

    def to_result(self):
        return Source(
            source_id=self.source_id,
            name=self.name,
            owner_email=self.owner_email,
            namespace=self.namespace.to_result(),
            category=self.category
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from data_pipeline.config import get_config
from data_pipeline.helpers.yelp_avro_store import _AvroStringStore
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer


logger = get_config().logger


class _AvroPayload(object):

    def __init__(
        self,
        schema_id,
        reader_schema_id=None,
        payload=None,
        payload_data=None,
        dry_run=False
    ):
        self._set_schema_id(schema_id)
        self._set_reader_schema_id(reader_schema_id)
        self._set_dry_run(dry_run)
        self._set_payload_or_payload_data(payload, payload_data)

    @property
    def _schematizer(self):
        return get_schematizer()

    @property
    def schema_id(self):
        return self._schema_id

    def _set_schema_id(self, schema_id):
        if not isinstance(schema_id, int):
            raise TypeError("Schema id should be an int")
        self._schema_id = schema_id

    @property
    def reader_schema_id(self):
        return self._reader_schema_id

    def _set_reader_schema_id(self, reader_schema_id):
        if (reader_schema_id is not None and
                not isinstance(reader_schema_id, int)):
            raise TypeError("Reader Schema id should be an int")
        self._reader_schema_id = reader_schema_id or self.schema_id

    @property
    def dry_run(self):
        return self._dry_run

    def _set_dry_run(self, dry_run):
        self._dry_run = dry_run

    def _set_payload_or_payload_data(self, payload, payload_data):
        # payload or payload_data are lazily constructed only on request
        is_not_none_payload = payload is not None
        is_not_none_payload_data = payload_data is not None

        if is_not_none_payload and is_not_none_payload_data:
            raise TypeError("Cannot pass both payload and payload_data.")
        if is_not_none_payload:
            self._set_payload(payload)
        elif is_not_none_payload_data:
            self._set_payload_data(payload_data)
        else:
            raise TypeError("Either payload or payload_data must be provided.")

    @property
    def payload(self):
        self._set_payload_if_necessary(self._payload_data)
        return self._payload

    def _set_payload(self, payload):
        if not isinstance(payload, bytes):
            raise TypeError("Payload must be bytes")
        self._payload = payload
        self._payload_data = None  # force payload_data to be re-decoded

    @property
    def payload_data(self):
        self._set_payload_data_if_necessary(self._payload)
        return self._payload_data

    @property
    def printable_payload_data(self):
        return self._get_printable_payload_data(self.payload_data)

    def _get_printable_payload_data(self, data):
        if not isinstance(data, dict):
            return data.encode('hex') if isinstance(data, bytes) else data
        return {
            key: self._get_printable_payload_data(value)
            for key, value in data.iteritems()
        }

    def _set_payload_data(self, payload_data):
        """We shold check to verify that payload data is not None
        payload_data should not necessarily be a dict. example if the schema is something like
        {
          "type": "fixed",
          "size": 16,
          "namespace": "yelp.data_pipeline",
          "name": "initialization_vector",
          "doc": "Serializes an initialization vector for encrypting PII."
        }

        then the corresponding payload data will be 1234123412341234 which is not a dict

        In ideal scenario we should be using avro.io.validate function to validate if
        the given payload_data is a valid datum, but this check happens anyway when we
        encode payload_data with avro_schema.
        """
        if payload_data is None:
            raise TypeError("Payload Data cannot be None")
        self._payload_data = payload_data
        self._payload = None  # force payload to be re-encoded

    def _set_payload_data_if_necessary(self, payload):
        if self._payload_data is None:
            self._payload_data = self._decode_payload(payload)

    def _set_payload_if_necessary(self, payload_data):
        if self._payload is None:
            self._payload = self._encode_payload_data(payload_data)

    def _encode_payload_data(self, payload_data):
        if self.dry_run:
            return repr(payload_data)
        return self._avro_string_writer.encode(
            message_avro_representation=payload_data
        )

    def _decode_payload(self, payload):
        return self._avro_string_reader.decode(
            encoded_message=payload
        )

    @property
    def _avro_string_writer(self):
        """get the writer from store if already exists"""
        return _AvroStringStore().get_writer(self.schema_id)

    @property
    def _avro_string_reader(self):
        """get the reader from store if already exists"""
        return _AvroStringStore().get_reader(
            reader_id_key=self.reader_schema_id,
            writer_id_key=self.schema_id
        )

    def reload_data(self):
        """Populate the payload data or the payload if it hasn't done so.
        """
        self._set_payload_data_if_necessary(self._payload)
        self._set_payload_if_necessary(self._payload_data)



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from enum import Enum

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel
from data_pipeline.schematizer_clientlib.models.source import _Source

"""
Represents the data of a refresh. A refresh represents the parameters
associated with an individual full refresh of a particular source.

Args:
    refresh_id (int): The id of the refresh.
    source (data_pipeline.schematizer_clientlib.models.source.Source):
        The source of the refresh.
    status (RefreshStatus): The current status of the refresh.
    offset (int): Last known offset that has been refreshed.
    batch_size (int): The number of rows to be refreshed per batch.
    priority (Priority): The priority of the refresh
    filter_condition (str): The filter_condition associated with the refresh.
    avg_rows_per_second_cap (int): The throughput throttling cap to be used when
        the refresh is run.
    created_at (str): The timestamp when the refresh is created in ISO-8601
        format.
    updated_at (str): The timestamp when the refresh is last updated in ISO-8601
        format.
"""
Refresh = namedtuple(
    'Refresh',
    [
        'refresh_id',
        'source',
        'status',
        'offset',
        'batch_size',
        'priority',
        'filter_condition',
        'avg_rows_per_second_cap',
        'created_at',
        'updated_at'
    ]
)


class Priority(Enum):
    LOW = 25
    MEDIUM = 50
    HIGH = 75
    MAX = 100


class RefreshStatus(Enum):
    NOT_STARTED = "NOT_STARTED"
    IN_PROGRESS = "IN_PROGRESS"
    PAUSED = "PAUSED"
    SUCCESS = "SUCCESS"
    FAILED = "FAILED"


class _Refresh(BaseModel):

    def __init__(
        self,
        refresh_id,
        source,
        status,
        offset,
        batch_size,
        priority,
        filter_condition,
        created_at,
        updated_at,
        # Has to go last since it's optional
        avg_rows_per_second_cap=None
    ):
        self.refresh_id = refresh_id
        self.source = source
        self.status = RefreshStatus[status]
        self.offset = offset
        self.batch_size = batch_size
        self.priority = Priority[priority]
        self.filter_condition = filter_condition
        self.avg_rows_per_second_cap = avg_rows_per_second_cap
        self.created_at = created_at
        self.updated_at = updated_at

    @classmethod
    def from_response(cls, response):
        return cls(
            refresh_id=response.refresh_id,
            source=_Source.from_response(response.source),
            status=response.status,
            offset=response.offset,
            batch_size=response.batch_size,
            priority=response.priority,
            filter_condition=response.filter_condition,
            avg_rows_per_second_cap=getattr(response, 'avg_rows_per_second_cap', None),
            created_at=response.created_at,
            updated_at=response.updated_at
        )

    def to_result(self):
        return Refresh(
            refresh_id=self.refresh_id,
            source=self.source.to_result(),
            status=self.status,
            offset=self.offset,
            batch_size=self.batch_size,
            priority=self.priority,
            filter_condition=self.filter_condition,
            avg_rows_per_second_cap=self.avg_rows_per_second_cap,
            created_at=self.created_at,
            updated_at=self.updated_at
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import threading

import simplejson
from cached_property import cached_property

from data_pipeline._clog_writer import ClogWriter
from data_pipeline.config import get_config
from data_pipeline.message import RegistrationMessage
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer

logger = get_config().logger


class Registrar(object):
    """This class holds the main functionality for Producer/Consumer registration

    Currently, the registrar holds an internal mapping of schema_ids to the last time
    they are used by the Client's subclass. Client will provide functionality for updating
    this mapping as it sends/receives messages. Every time the given threshold time is
    reached, the Client will send a serialized message using clog.

    Args:
        team_name (str): Team name, as defined in `sensu_handlers::teams` (see y/sensu-teams).
        client_name (str): name of the associated client.
        client_type (str): type of the client the _Registrar is associated to.
            Could be either producer or consumer.
        expected_frequency_seconds (int): How frequently, in seconds, that the client expects
            to run to produce or consume messages.
        threshold (int): The amount of time that should elapse in between the client sending
            registration messages (seconds).
    """

    # Default period for sending registration msg is 10 min
    DEFAULT_REGISTRATION_THRESHOLD_SECONDS = 600

    def __init__(
        self,
        team_name,
        client_name,
        client_type,
        expected_frequency_seconds,
        threshold=DEFAULT_REGISTRATION_THRESHOLD_SECONDS
    ):
        self.team_name = team_name
        self.client_name = client_name
        self.client_type = client_type
        self.threshold = threshold
        self.send_messages = False
        self.schema_to_last_seen_time_map = {}
        self.expected_frequency_seconds = expected_frequency_seconds
        self.clog_writer = ClogWriter()
        self.current_thread = None

    def publish_registration_messages(self):
        """
        Publish all registration messages to Scribe.
        """
        for message in self.get_registration_messages():
            self.clog_writer.publish(message)

    def get_registration_messages(self):
        """
        This function will convert the internal state of the registrar into instances of
        the RegistrationMessage class.
        """
        registration_messages = [
            self._create_registration_message(schema_id, timestamp)
            for schema_id, timestamp in self.schema_to_last_seen_time_map.iteritems()
        ]
        return registration_messages

    def _create_registration_message(self, schema_id, last_seen_timestamp):
        """
        Return single instance of a RegistrationMessage using passed in schema_id and
        timestamp.
        """
        payload_data = self._registration_message_payload(schema_id, last_seen_timestamp)
        return RegistrationMessage(
            schema_id=self.registration_schema.schema_id,
            payload_data=payload_data
        )

    def _registration_message_payload(self, schema_id, last_seen_timestamp):
        return {
            "team_name": self.team_name,
            "client_name": self.client_name,
            "client_type": self.client_type,
            "timestamp": last_seen_timestamp,
            "expected_frequency_seconds": self.expected_frequency_seconds,
            "schema_id": schema_id
        }

    @cached_property
    def registration_schema(self):
        schema_json = self._registration_schema
        return get_schematizer().register_schema(
            namespace=schema_json['namespace'],
            source=schema_json['name'],
            schema_str=simplejson.dumps(schema_json),
            source_owner_email='bam+data_pipeline@yelp.com',
            contains_pii=False
        )

    @cached_property
    def _registration_schema(self):
        schema_file = os.path.join(
            os.path.dirname(__file__),
            'schemas/registration_message_v1.avsc'
        )
        with open(schema_file, 'r') as f:
            schema_string = f.read()
        return simplejson.loads(schema_string)

    def register_tracked_schema_ids(self, schema_id_list):
        """This function is used to specify the list of avro schema IDs that this Client
            will use. When called it, it will immediately publish registration messages.

        Args:
            schema_id_list (list[int]): List of the schema IDs that the client will use.
        """
        for schema_id in schema_id_list:
            self.schema_to_last_seen_time_map[schema_id] = None
        self.publish_registration_messages()

    def update_schema_last_used_timestamp(self, schema_id, timestamp_in_milliseconds):
        """
        This function updates the last time that the given schema_id was used to value
        timestamp if the given timestamp occurred more recently than the last time the
        schema_id was used.

        Usage:
            This function is called by the Client subclass whenever it receives a message.

        Args:
            schema_id (int): Schema IDs of the message the Client received.
            timestamp_in_milliseconds (long): The utc time of the message that the Client
                                              received in milliseconds.
        """
        current_timestamp = self.schema_to_last_seen_time_map.get(schema_id)
        if current_timestamp is None or timestamp_in_milliseconds > current_timestamp:
            self.schema_to_last_seen_time_map[schema_id] = timestamp_in_milliseconds

    def start(self):
        """Start periodically sending registration messages after threshold amount of time"""
        if not self.send_messages:
            self.send_messages = True
            self.current_thread = threading.Timer(self.threshold, self._wake)
            self.current_thread.start()

    def stop(self):
        """Force Client to stop periodically sending registration messages"""
        self.send_messages = False
        if self.current_thread:
            self.current_thread.cancel()
        # Send registration messages when the Registrar is stopped
        self.publish_registration_messages()

    def _wake(self):
        """This class periodically sends registration messages using Clog"""
        if self.send_messages:
            self.publish_registration_messages()
            # The purpose of the Timer is for _wake to ensure it is called
            self.current_thread = threading.Timer(self.threshold, self._wake)
            self.current_thread.start()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.data_source_type_enum import DataSourceTypeEnum
from data_pipeline.schematizer_clientlib.models.model_base import BaseModel


"""
Represent the data of the mapping between a data source and a consumer group.
A data source represents a namespace or a source.

Args:
    consumer_group_data_source_id (int): The id of the mapping between consumer
        group and data source.
    consumer_group_id (str): The id of the consumer group.
    data_source_type
    (data_pipeline.schematizer_clientlib.models.data_source_type_enum.DataSourceTypeEnum):
        The type of the data_source.
    data_source_id: The id of the data target.  Depending on the data source
        type, it may be a namespace id or source id.
"""
ConsumerGroupDataSource = namedtuple(
    'ConsumerGroupDataSource',
    ['consumer_group_data_source_id', 'consumer_group_id', 'data_source_type',
     'data_source_id']
)


class _ConsumerGroupDataSource(BaseModel):
    """Internal class used to convert from/to various data structure and
    facilitate constructing the return value of schematizer functions.
    """

    def __init__(self, consumer_group_data_source_id, consumer_group_id,
                 data_source_type, data_source_id):
        self.consumer_group_data_source_id = consumer_group_data_source_id
        self.consumer_group_id = consumer_group_id
        self.data_source_type = data_source_type
        self.data_source_id = data_source_id

    @classmethod
    def from_response(cls, response):
        return cls(
            consumer_group_data_source_id=response.consumer_group_data_source_id,
            consumer_group_id=response.consumer_group_id,
            data_source_type=DataSourceTypeEnum[response.data_source_type],
            data_source_id=response.data_source_id
        )

    def to_result(self):
        return ConsumerGroupDataSource(
            consumer_group_data_source_id=self.consumer_group_data_source_id,
            consumer_group_id=self.consumer_group_id,
            data_source_type=self.data_source_type,
            data_source_id=self.data_source_id
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import re
import signal
import time
from inspect import getmembers
from optparse import OptionGroup
from uuid import UUID

import simplejson
from kafka import KafkaClient
from kafka_utils.util import offsets
from yelp_batch.batch import Batch
from yelp_batch.batch import batch_command_line_options
from yelp_batch.batch import batch_configure

from data_pipeline import __version__
from data_pipeline._fast_uuid import FastUUID
from data_pipeline.base_consumer import ConsumerTopicState
from data_pipeline.config import get_config
from data_pipeline.consumer import Consumer
from data_pipeline.expected_frequency import ExpectedFrequency
from data_pipeline.message import Message
from data_pipeline.schematizer_clientlib.schematizer import get_schematizer
from data_pipeline.servlib.config_util import load_default_config
from data_pipeline.tools.timestamp_to_offset_mapper import get_first_offset_at_or_after_start_timestamp

logger = get_config().logger


class Tailer(Batch):
    """Tailer subscribes to (a) Kafka topic(s) and logs any messages pushed to these,
    applying a transformation on the Avro binary data such that it is readable."""
    enable_error_emails = False

    @property
    def version(self):
        """Overriding this so we'll get the clientlib version number when
        the tailer is run with --version.
        """
        return "data_pipeline {}".format(__version__)

    @batch_command_line_options
    def _define_tailer_options(self, option_parser):
        opt_group = OptionGroup(
            option_parser,
            'Data Pipeline Tailer options'
        )

        opt_group.add_option(
            '--topic',
            type='string',
            action='append',
            dest='topics',
            default=[],
            help=(
                'The topic to tail.  Can be specified multiple times to tail '
                'more than one topic.  Topics can optionally include an offset '
                'by formatting the topic like "topic_name|offset".  If offset '
                'is not specified, its value will come from '
                '--offset-reset-location.'
            ),
        )
        opt_group.add_option(
            '--cluster-name',
            type='string',
            help=(
                'The cluster_name (region) of kafka cluster to connect to. '
                'All topics should belong to the same kafka cluster name. '
                'Defaults to local region where the tailer is running. '
                '(Example: uswest1-devc, uswest2-prod)'
            ),
        )
        opt_group.add_option(
            '--config-file',
            help=(
                'If set, will use the provided configuration file to setup '
                'the data pipeline tools. '
                '(Default is %default)'
            ),
            default='/nail/srv/configs/data_pipeline_tools.yaml'
        )
        opt_group.add_option(
            '--env-config-file',
            help=(
                'If set, will use the provided configuration file as the env '
                'overrides file to setup the data pipeline tools. '
                '(Default is %default)'
            ),
            default=None
        )
        opt_group.add_option(
            '--namespace',
            type='string',
            help=(
                'If given, the namespace and source will be used to lookup '
                'topics, which will be added to any topics otherwise provided. '
                '(Example: refresh_primary.yelp)'
            )
        )
        opt_group.add_option(
            '--source',
            type='string',
            help=(
                'If given, the namespace and source will be used to lookup '
                'topics, which will be added to any topics otherwise provided. '
                '(Example: business)'
            )
        )
        opt_group.add_option(
            '--only-newest',
            action="store_true",
            help=(
                "If given, will limit the topics retrieved from namespace and source to the "
                "topic that has been most recently updated"
            )
        )
        opt_group.add_option(
            '--offset-reset-location',
            type='string',
            default='largest',
            help=(
                'Controls where the Kafka tailer starts from.  Set to smallest '
                'to consume topics from the beginning.  Set to largest to '
                'consume topics from the tail. '
                '(Default is %default)'
            )
        )
        opt_group.add_option(
            '--message-limit',
            type='int',
            default=None,
            help=(
                'Provide a message limit to quit when the specified number of '
                'messages have been output.'
            )
        )
        opt_group.add_option(
            '-f', '--field',
            type='string',
            action='append',
            dest='fields',
            default=['payload_data'],
            help=(
                'The fields of the message to output (default: %default). '
                'The following fields '
                'are available: {}'.format(
                    ', '.join(self._public_message_field_names)
                )
            ),
        )
        opt_group.add_option(
            '--all-fields',
            default=False,
            action='store_true',
            help=(
                'If set, all fields of the message will be output. This is a '
                'shortcut to doing --field for every field name.'
            )
        )
        opt_group.add_option(
            '--json',
            default=False,
            action='store_true',
            help=(
                'If set, the output is converted from into json.'
            )
        )
        opt_group.add_option(
            '--iso-time',
            default=False,
            action='store_true',
            help=(
                'If set, the output for time fields will be in ISO 8601 '
                'format rather than epoch timestamp.'
            )
        )
        opt_group.add_option(
            '--end-timestamp',
            default=None,
            type=int,
            help=(
                'If set, the tailer will only output messages up to the given end date. '
                'Formatted using epoch timestamp.'
            )
        )
        opt_group.add_option(
            '--start-timestamp',
            default=None,
            type=int,
            help=(
                'If set, will set starting offset to be the offset of the first available'
                ' message that comes after or at the given start-timestamp. If a starting offset is'
                ' manually set using the --topic option, then this will not override it.'
                ' Formatted using epoch timestamp'
            )
        )
        return opt_group

    @property
    def _all_message_field_names(self):
        return [
            name for name, value in getmembers(
                Message,
                lambda v: isinstance(v, property)
            )
        ]

    @property
    def _public_message_field_names(self):
        return [
            name for name in self._all_message_field_names
            if not name.startswith('_') and name not in [
                # these fields contain redundant information and as byte-fields
                # can result in UnicodeDecodeError from the simplejson encoder
                'payload',
                'previous_payload',
                'uuid',

                # avro_repr contains redundant information
                'avro_repr',

                # upstream_position_info is for internal use, should always
                # be 'None' to the messages the Tailer will see
                'upstream_position_info',

                # dry_run is always going to be 'False' to messages the Tailer
                # would see
                'dry_run'
            ]
        ]

    @batch_configure
    def _configure_tools(self):
        load_default_config(
            self.options.config_file,
            self.options.env_config_file
        )

        # We setup logging 'early' since we want it available for setup_topics
        self._setup_logging()

        self.kafka_client = KafkaClient(get_config().cluster_config.broker_list)

        self._setup_topics()
        if len(self.topic_to_offsets_map) == 0:
            self.option_parser.error("At least one topic must be specified.")

        if self.options.start_timestamp is not None and self.options.start_timestamp >= int(time.time()):
            self.option_parser.error("--start-timestamp should not be later than current time")

        if self.options.start_timestamp is not None and self.options.end_timestamp and (
            self.options.start_timestamp > self.options.end_timestamp
        ):
            self.option_parser.error("--end-timestamp must not be smaller than --start-timestamp")

        if self.options.all_fields:
            self.options.fields = self._public_message_field_names

        self._verify_offset_ranges()

    def _verify_offset_ranges(self):
        """This is to clarify and enforce only using offsets inside of our actual offset range to avoid
        confusing errors such as those found in DATAPIPE-628"""
        topic_to_partition_offset_map = {
            topic: None if consumer_topic_state is None
            else consumer_topic_state.partition_offset_map
            for topic, consumer_topic_state in self.topic_to_offsets_map.items()
        }
        # If we import get_topics_watermarks directly from offsets, then mock will not properly patch it in testing.
        watermarks = offsets.get_topics_watermarks(
            self.kafka_client,
            topic_to_partition_offset_map,
            # We do not raise on error as we do this verification later on and we
            # want to keep the error message clear
            raise_on_error=False
        )
        for topic, partition_offset_map in topic_to_partition_offset_map.iteritems():
            if partition_offset_map is not None:
                for partition, offset in partition_offset_map.iteritems():
                    highmark = watermarks[topic][partition].highmark
                    lowmark = watermarks[topic][partition].lowmark
                    if offset < lowmark or offset > highmark:
                        self.option_parser.error(
                            "Offset ({}) for topic: {} (partition: {}) is out of range ({}-{})".format(
                                offset,
                                topic,
                                partition,
                                lowmark,
                                highmark
                            )
                        )

    def _setup_topics(self):
        self.topic_to_offsets_map = {}
        self._setup_manual_topics()
        self._setup_schematizer_topics()
        if self.options.start_timestamp is not None:
            self._setup_start_timestamp_topics(self.options.start_timestamp)

    def _setup_manual_topics(self):
        for topic in self.options.topics:
            offset = None

            # https://regex101.com/r/fL0eD9/3
            match = re.match('^(.*)\|(\d+)$', topic)
            if match:
                topic = match.group(1)
                offset = ConsumerTopicState({0: int(match.group(2))}, None)

            self.topic_to_offsets_map[str(topic)] = offset

    def _setup_schematizer_topics(self):
        if self.options.namespace or self.options.source:
            schematizer = get_schematizer()
            additional_topics = schematizer.get_topics_by_criteria(
                namespace_name=self.options.namespace,
                source_name=self.options.source
            )
            if self.options.only_newest:
                additional_topics = self._filter_by_most_recently_updated(additional_topics)
            logger.info(
                "Received {} new topics from --source and --namespace options".format(
                    len(additional_topics)
                )
            )
            for topic in additional_topics:
                if str(topic.name) not in self.topic_to_offsets_map:
                    self.topic_to_offsets_map[str(topic.name)] = None

    def _filter_by_most_recently_updated(self, topics):
        if not topics:
            return []
        sorted_topics = sorted(topics, key=lambda topic: topic.updated_at)
        return [sorted_topics[-1]]

    def _setup_start_timestamp_topics(self, start_timestamp):
        """Sets the offsets of all topics with no offset to be that topic's first offset
        that comes after --start-timestamp"""
        no_offset_topics = [
            topic
            for topic, offset in self.topic_to_offsets_map.iteritems()
            if offset is None
        ]
        logger.info(
            "Getting starting offsets for {} based on --start-timestamp".format(no_offset_topics)
        )
        start_timestamp_topic_to_offset_map = get_first_offset_at_or_after_start_timestamp(
            self.kafka_client,
            no_offset_topics,
            start_timestamp
        )
        for topic, consumer_topic_state in start_timestamp_topic_to_offset_map.iteritems():
            self.topic_to_offsets_map[topic] = start_timestamp_topic_to_offset_map[topic]

    @batch_configure
    def _configure_signals(self):
        self._running = True

        def handle_signal(signum, frame):
            self._running = False
        signal.signal(signal.SIGTERM, handle_signal)
        signal.signal(signal.SIGINT, handle_signal)

    def run(self):
        logger.info(
            "Starting to consume from {}".format(self.topic_to_offsets_map)
        )

        with Consumer(
            # The tailer name should be unique - if it's not, partitions will
            # be split between multiple tailer instances
            'data_pipeline_tailer-{}'.format(
                str(UUID(bytes=FastUUID().uuid4()).hex)
            ),
            'bam',
            ExpectedFrequency.constantly,
            self.topic_to_offsets_map,
            auto_offset_reset=self.options.offset_reset_location,
            cluster_name=self.options.cluster_name
        ) as consumer:
            message_count = 0
            while self.keep_running(message_count):
                message = consumer.get_message(blocking=True, timeout=0.1)
                if message is not None:
                    if self.options.end_timestamp is None or message.timestamp < self.options.end_timestamp:
                        print self._format_message(message)
                        message_count += 1
                    else:
                        self._running = False
                        logger.info(
                            "Latest message surpasses --end-timestamp. Stopping tailer..."
                        )

    def _get_message_result_dict(self, message):
        return {
            field: getattr(message, field)
            for field in self.options.fields if hasattr(message, field)
        }

    def _walk_dict(self, node, transform_item):
        for key, item in node.items():
            if isinstance(item, dict):
                self._walk_dict(item, transform_item)
            else:
                node[key] = transform_item(key, item)

    def _iso_time(self, result_dict):
        def _transform_item_iso_time(key, item):
            if key.startswith('time') and isinstance(item, int):
                return time.strftime(
                    '%Y-%m-%dT%H:%M:%S',
                    time.gmtime(item)
                )
            return item
        self._walk_dict(result_dict, _transform_item_iso_time)

    def _format_message(self, message):
        result_dict = self._get_message_result_dict(message)
        if self.options.iso_time:
            self._iso_time(result_dict)
        if self.options.json:
            return simplejson.dumps(
                obj=result_dict,
                sort_keys=True,

                # Objects can use _asdict() to be encoded as JSON objects
                namedtuple_as_object=True,

                # Use an object's __repr__() to return a serializable version
                # of an object, rather than raising a TypeError, if the object
                # does not define an _asdict() method
                default=lambda x: repr(x)
            )
        else:
            return result_dict

    def keep_running(self, message_count):
        return self._running and (
            self.options.message_limit is None or
            message_count < self.options.message_limit
        )


if __name__ == '__main__':
    Tailer().start()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import simplejson

from data_pipeline.tools.introspector.base_command import IntrospectorCommand
from data_pipeline.tools.introspector.models import IntrospectorSource


class SourceInfoCommand(IntrospectorCommand):
    @classmethod
    def add_parser(cls, subparsers):
        info_command_parser = subparsers.add_parser(
            "source",
            description="Get information on a specific data pipeline source.",
            add_help=False
        )

        info_command_parser.add_argument(
            '--active-sources',
            default=False,
            action='store_true',
            help=(
                'If set, this command will also return information about active '
                'topics for this source. '
                'This is a time expensive operation.'
            )
        )

        cls.add_base_arguments(info_command_parser)
        cls.add_source_and_namespace_arguments(info_command_parser)

        info_command_parser.set_defaults(
            command=lambda args: cls("data_pipeline_instropsector_info_source").run(
                args,
                info_command_parser
            )
        )

    def process_args(self, args, parser):
        super(SourceInfoCommand, self).process_args(args, parser)
        self.process_source_and_namespace_args(args, parser)

    def info_source(
        self,
        source_id=None,
        source_name=None,
        namespace_name=None,
        active_sources=False
    ):
        info_source = None
        if source_id:
            info_source = self.schematizer.get_source_by_id(source_id)
        else:
            sources = self.schematizer.get_sources_by_namespace(namespace_name)
            for source in sources:
                if source.name == source_name:
                    info_source = source
                    break
            if not info_source:
                raise ValueError("Given SOURCE_NAME|NAMESPACE_NAME doesn't exist")
        info_source = IntrospectorSource(
            info_source
        ).to_ordered_dict()
        topics = self.list_topics(
            source_id=info_source["source_id"]
        )
        if active_sources:
            info_source['active_topic_count'] = len(
                [topic for topic in topics if topic['message_count']]
            )
        info_source['topics'] = topics
        return info_source

    def run(self, args, parser):
        self.process_args(args, parser)
        print simplejson.dumps(
            self.info_source(
                source_id=self.source_id,
                source_name=self.source_name,
                namespace_name=self.namespace,
                active_sources=args.active_sources
            )
        )



# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import optparse
import re

import simplejson as json
import yelp_batch
from cached_property import cached_property
from yelp_batch.batch import batch_command_line_options
from yelp_batch.batch import os

from data_pipeline.tools._glob_util import get_file_paths_from_glob_patterns

# See https://regex101.com/r/kC0kZ1/2
CREATE_TABLE_REGEX = re.compile('^create(\s*table)?\s*((.+)\.)?(\w+)\s*\(?')

# See https://regex101.com/r/zG9kV1/2
PRIMARY_KEY_REGEX = re.compile('^primary\s*key\s*\((.+)?\)')

# See https://regex101.com/r/kD8iN5/17
FIELD_LINE_REGEX = re.compile(
    '^(\w+)\s*(\w+)\s*(\(\s*(\d+|\d+\s*\,\s*\d+)\s*\))?\s*(?P<pk>primary\s+key)?\s*(not\s+null|null)?\s*((default)\s+(\"|\')?(null|false|true|\d+\.\d+|\d+|[\w\s]*)(\"|\')?)?(\"|\')?.*,'  # noqa
)

# See https://regex101.com/r/bN3xL0
START_FIELDS_REGEX = re.compile('^.*\(')

# See https://regex101.com/r/bR7bH2
STOP_FIELDS_REGEX = re.compile('^\)')

REDSHIFT_SQL_TO_AVRO_TYPE_MAPPING = {
    'bigint': 'long',
    'bool': 'boolean',
    'boolean': 'boolean',
    'bpchar': 'string',
    'char': 'string',
    'character': 'string',
    'date': 'string',
    'decimal': 'double',
    'numeric': 'double',
    'double': 'double',
    'float': 'double',
    'float4': 'float',
    'float8': 'double',
    'int': 'int',
    'int2': 'int',
    'int4': 'int',
    'int8': 'long',
    'integer': 'int',
    'nchar': 'string',
    'nvarchar': 'string',
    'real': 'float',
    'smallint': 'int',
    'text': 'string',
    'timestamp': 'long',
    'varchar': 'string'
}


def _sanitize_line(line):
    return line.strip().lower()


class RedshiftFieldLineToAvroFieldConverter(object):
    """ Converter for a single redshift column definition line in a
    `CREATE TABLE` statement.

    This should eventually be replaced by DATAPIPE-353.
    """

    def __init__(self, field_line, pkeys):
        """
        Args:
            field_line(string): Content of a column definition line from a
                redshift *.sql file
            pkeys([string]): A list of the primary keys, used for determining
                the meta attribute of "pkey"
        """
        self.field_line = _sanitize_line(field_line)
        self.pkeys = pkeys

    @cached_property
    def avro_field(self):
        field = {
            "name": self.name,
            "type": self.avro_type,
            "doc": ""
        }
        field.update(self.avro_meta_attributes)
        return field

    @cached_property
    def name(self):
        return self._regex_matcher.group(1)

    @cached_property
    def avro_core_type(self):
        return REDSHIFT_SQL_TO_AVRO_TYPE_MAPPING[self.sql_type]

    @cached_property
    def avro_type(self):
        avro_type = self.avro_core_type
        if self.nullable:
            if self.default_null:
                return ['null', avro_type]
            else:
                return [avro_type, 'null']
        else:
            return avro_type

    @cached_property
    def sql_type(self):
        return self._regex_matcher.group(2)

    @cached_property
    def sql_default(self):
        """ Return the default value defined for the column, if any.

            Note:
                This will succeed only if the 'default' follows the 'NOT NULL'/
                'NULL' on the column line. I've reached the limits of what
                black magic I'm willing to deal with in this regex and
                DATAPIPE-353 should be replacing this eventually anyway. :)
        """
        return self._regex_matcher.group(10)

    @cached_property
    def nullable(self):
        nullable_str = self._regex_matcher.group(6)
        return not(nullable_str and re.search('^(not\s+null)', nullable_str))

    @cached_property
    def default_null(self):
        return self.nullable and self.sql_default in ['null', None]

    @cached_property
    def avro_meta_attributes(self):
        meta = {}
        field_name = self.name
        for index, pkey_name in enumerate(self.pkeys):
            if pkey_name == field_name:
                meta['pkey'] = index + 1
                break
        if self.sql_type in ['varchar', 'nvarchar', 'text']:
            meta['maxlen'] = self.sql_type_width
        if self.sql_type in ['char', 'character', 'nchar', 'bpchar']:
            meta['fixlen'] = self.sql_type_width
        if self.sql_type in ['date', 'timestamp']:
            meta[self.sql_type] = True
        if self.sql_type in ['decimal', 'numeric']:
            meta['fixed_pt'] = True
            meta['precision'] = self.sql_type_width[0]
            meta['scale'] = self.sql_type_width[1]
        if self.default_null:
            meta['default'] = None
        elif self.sql_default is not None:
            if self.avro_core_type == 'boolean':
                if self.sql_default == 'true':
                    meta['default'] = True
                elif self.sql_default == 'false':
                    meta['default'] = False
                else:
                    try:
                        meta['default'] = bool(int(self.sql_default))
                    except ValueError:
                        # suppress the exception
                        pass
            elif self.avro_core_type in ['long', 'int']:
                try:
                    meta['default'] = int(self.sql_default)
                except ValueError:
                    # suppress the exception. This can be thrown when the
                    # default is something like 'getdate()'
                    pass
            elif self.avro_core_type in ['double', 'float']:
                try:
                    meta['default'] = float(self.sql_default)
                except ValueError:
                    # suppress the exception.
                    pass
            else:
                meta['default'] = self.sql_default
        return meta

    @cached_property
    def sql_type_width(self):
        """ Return the sql type width, which is an int defining the the
        maximum size for character types and a (presumably two element) list of
        ints (the precision and scale) for the decimal type.

        Note:
            Some redshift sql types have default widths associated to them, see
            http://docs.aws.amazon.com/redshift/latest/dg/r_Character_types.html
            for more details
        """
        width = self._regex_matcher.group(4)
        if width:
            if ',' in width:
                return [
                    int(part.strip())
                    for part in width.split(',')
                ]
            else:
                return int(width)
        else:
            if self.sql_type in ['text', 'bpchar', 'varchar', 'nvarchar']:
                return 256
            if self.sql_type in ['char', 'character', 'nchar']:
                return 1
            return None

    @cached_property
    def _regex_matcher(self):
        return FIELD_LINE_REGEX.search(self.field_line)


class RedshiftSQLToAVSCConverter(object):
    """ Simple converter from redshift *.sql CREATE TABLE definitions (such
    as those in yelp-main/schema/yelp_dw_redshift/tables) to data pipeline
    format Avro *.avsc schemas.

    This should eventually be replaced by DATAPIPE-353.

    Notes:
        This makes a number of assumptions about the input content, namely
        that there is a column definition per line, that is followed by
        convention in all yelp *.sql files - however this is NOT a general
        purpose parser/converter.
    """

    def __init__(self, sql_content, base_namespace, default_schema='public'):
        """
        Args:
            sql_content(string): Content of a redshift *.sql file
            base_namespace(string): The base namespace (the namespace will be
                a combination of "{base_namespace}.{schema}"
            default_schema(string): The default schema, for any tables
                encountered which do not specify a schema.
        """
        self.sql_content = sql_content
        self.base_namespace = base_namespace
        self.default_schema = default_schema

    @cached_property
    def avro_record(self):
        """ Get the data pipeline format Avro representation of
        self.sql_content.
        """
        return {
            'type': 'record',
            'namespace': self.namespace,
            'name': self.table,
            'doc': '',
            'pkey': self.pkeys,
            'fields': [
                field_line_converter.avro_field
                for field_line_converter in self.field_line_converters
            ]
        }

    @cached_property
    def namespace(self):
        return '{0}.{1}'.format(self.base_namespace, self.schema)

    @cached_property
    def schema(self):
        m = CREATE_TABLE_REGEX.search(self.create_table_line)
        return m.group(3) if m.group(3) else self.default_schema

    @cached_property
    def table(self):
        m = CREATE_TABLE_REGEX.search(self.create_table_line)
        if m.group(4):
            return m.group(4)
        else:
            raise ValueError("Could not locate the table name")

    @cached_property
    def sql_lines(self):
        return [_sanitize_line(line) for line in self.sql_content.split('\n')]

    @cached_property
    def create_table_line(self):
        for line in self.sql_lines:
            if CREATE_TABLE_REGEX.search(line):
                return line
        raise ValueError("Could not locate a 'CREATE TABLE' statement!")

    @cached_property
    def pkeys(self):
        pkeys = []
        # loop through field lines to extract primary keys
        for line in self.sql_lines:
            if self._get_primary_key_in_field_line(line):
                pkeys.append(self._get_primary_key_in_field_line(line))

        if self.primary_key_line:
            pkeys.extend([
                pkey.strip() for pkey in
                PRIMARY_KEY_REGEX.search(
                    self.primary_key_line
                ).group(1).split(',')
            ])
        return pkeys

    @cached_property
    def primary_key_line(self):
        for line in self.sql_lines:
            if self._is_primary_key_line(line):
                return line

    def _is_primary_key_line(self, line):
        return bool(PRIMARY_KEY_REGEX.search(line))

    def _get_primary_key_in_field_line(self, line):
        field_line = FIELD_LINE_REGEX.search(line)
        if field_line and field_line.group(5) is not None:
            # if primary key present in sql field line return field name
            return field_line.group(1)

    @cached_property
    def field_line_converters(self):
        return [
            RedshiftFieldLineToAvroFieldConverter(
                field_line=line,
                pkeys=self.pkeys
            )
            for line in self._raw_field_lines
        ]

    @cached_property
    def _raw_field_lines(self):
        raw_field_lines = []
        for line in self.sql_lines[self._find_field_lines_start_index():]:
            line = _sanitize_line(line=line)
            if self._is_stop_line(line=line):
                break
            elif FIELD_LINE_REGEX.search(line):
                raw_field_lines.append(line)
        return raw_field_lines

    def _find_field_lines_start_index(self):
        for index, line in enumerate(self.sql_lines):
            line = _sanitize_line(line=line)
            if self._is_start_line(line=line):
                return index

    def _is_start_line(self, line):
        return bool(START_FIELDS_REGEX.search(line))

    def _is_stop_line(self, line):
        return STOP_FIELDS_REGEX.search(line) or self._is_primary_key_line(line)


class RedshiftSQLToAVSCBatch(yelp_batch.batch.Batch):

    notify_emails = ['bam+batch@yelp.com']

    @batch_command_line_options
    def parse_options(self, option_parser):
        opt_group = optparse.OptionGroup(
            option_parser,
            "RedshiftSQLToAVSC Options"
        )
        opt_group.add_option(
            '--glob',
            action='append',
            type='string',
            default=[],
            dest='globs',
            help='[REQUIRED] Either a path to a specific CREATE TABLE redshift'
                 ' *.sql file, or a glob pattern for a directory containing '
                 'such files. (For example: '
                 '"/nail/home/USER/some_dw_redshift_tables/*.sql") '
                 'Note --glob may be provided multiple times.'
        )
        opt_group.add_option(
            '--base-namespace',
            type='string',
            default='yelp_dw_redshift',
            help='[REQUIRED] Base of the namespace. The namespace will be a '
                 'combination of "{base-namespace}.{schema}" and it is best to '
                 'choose a base-namespace which reflects the data store '
                 'associated with the table (such as yelp_dw_redshift for the '
                 'yelp datawarehouse redshift tables). '
                 'Default is "%default"'
        )
        opt_group.add_option(
            '--default-schema',
            type='string',
            default='public',
            help='[REQUIRED] default schema for tables without any specified. '
                 'The namespace will be a combination of '
                 '"{base-namespace}.{schema}". '
                 'Default is "%default"'
        )
        opt_group.add_option(
            '--overwrite',
            action="store_true",
            default=False,
            help='Overwrite existing *.avsc files with new output from the '
                 'conversion run. '
                 'Default is "%default"'
        )
        return opt_group

    def run(self):
        """ Primary entry point for the batch
        """
        sql_file_paths = get_file_paths_from_glob_patterns(
            glob_patterns=self.options.globs
        )
        for sql_file_path in sql_file_paths:
            avsc_file_path = sql_file_path.replace('.sql', '.avsc')
            self.log.info(
                'Converting "{0}" to "{1}"'.format(
                    sql_file_path,
                    avsc_file_path
                )
            )
            if os.path.exists(avsc_file_path) and not self.options.overwrite:
                self.log.info(
                    'Skipping "{0}", use "--overwrite" to overwrite existing '
                    '*.avsc files.'.format(
                        avsc_file_path
                    )
                )
                continue
            self.convert_sql_to_avsc(
                avsc_file_path=avsc_file_path,
                sql_file_path=sql_file_path
            )

    def convert_sql_to_avsc(self, avsc_file_path, sql_file_path):
        with open(sql_file_path) as sql_file:
            sql_content = sql_file.read()
        converter = RedshiftSQLToAVSCConverter(
            sql_content=sql_content,
            base_namespace=self.options.base_namespace,
            default_schema=self.options.default_schema
        )
        avro = converter.avro_record
        with open(avsc_file_path, 'w') as avsc_file:
            self.log.info('Writing "{0}"'.format(avsc_file_path))
            json.dump(
                obj=avro,
                fp=avsc_file,
                indent='    ',
                sort_keys=True
            )


if __name__ == "__main__":
    RedshiftSQLToAVSCBatch().start()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from kafka_utils.util.offsets import get_topics_watermarks


def get_actual_published_messages_count(
    kafka_client,
    topics,
    topic_tracked_offset_map,
    raise_on_error=True,
):
    """Get the actual number of published messages of specified topics.

    Args:
        kafka_client (kafka.client.KafkaClient): kafka client
        topics ([str]): List of topic names to get message count
        topic_tracked_offset_map (dict(str, int)): dictionary which
            contains each topic and its current stored offset value.
        raise_on_error (Optional[bool]): if False,  the function ignores
            missing topics and missing partitions. It still may fail on
            the request send.  Default to True.

    Returns:
        dict(str, int): Each topic and its actual published messages count
            since last offset.  If a topic or partition is missing when
            `raise_on_error` is False, the returned dict will not contain
            the missing topic.

    Raises:
        :class:`~yelp_kafka.error.UnknownTopic`: upon missing topics and
            raise_on_error=True
        :class:`~yelp_kafka.error.UnknownPartition`: upon missing partitions
        and raise_on_error=True
        FailedPayloadsError: upon send request error.
    """
    topic_watermarks = get_topics_watermarks(
        kafka_client,
        topics,
        raise_on_error=raise_on_error
    )

    topic_to_published_msgs_count = {}
    for topic, partition_offsets in topic_watermarks.iteritems():
        high_watermark = partition_offsets[0].highmark
        offset = topic_tracked_offset_map.get(topic, 0)
        topic_to_published_msgs_count[topic] = high_watermark - offset

    return topic_to_published_msgs_count



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import glob
import inspect
import json
import optparse
import sys
from os import path
from pprint import pformat
from socket import gethostname

import yelp_batch
from yelp_batch.batch import batch_command_line_options

from data_pipeline.config import get_config
from data_pipeline.helpers.frozendict_json_encoder import FrozenDictEncoder
from data_pipeline.tools._glob_util import get_file_paths_from_glob_patterns
from data_pipeline.tools.schema_ref import SchemaRef


class FileBootstrapperBase(object):

    log = get_config().logger

    def __init__(
            self,
            schema_ref,
            file_paths,
            override_metadata,
            file_extension=None
    ):
        """
        Args:
            schema_ref(SchemaRef): SchemaRef to use for looking up metadata
            file_paths(set([str])): A list of file paths to use for bootstrapping
            override_metadata(boolean): If True then existing metadata (such as
                notes, categories, etc) will be overwritten with provided
                schema_ref, otherwise existing metadata will be preserved.
            file_extension(str): Must be specified by subclasses, this should be
                string specifying the file extension the subclass operates on,
                for example 'sql' or 'avsc'
        """
        self.api = get_config().schematizer_client
        self.log = get_config().logger
        self.schema_ref = schema_ref
        self.override_metadata = override_metadata
        self.file_extension = file_extension
        self.file_paths = set([
            file_path
            for file_path in file_paths
            if self.is_correct_file_extension(file_path)
        ])

    def register_file(self, file_path):
        """ Register a file specified and return the schema
        registration results. Subclasses must implement this to handle
        their own file type appropriately.
        """
        raise NotImplementedError("Subclasses should implement this!")

    def is_correct_file_extension(self, file_path):
        return file_path.endswith('.{}'.format(self.file_extension))

    def logged_register_file(self, file_path):
        self.log.info('-' * 40)
        self.log.info('- Registering *.{0} file: {1}'.format(
            self.file_extension,
            file_path
        ))
        return self.register_file(file_path)

    def bootstrap_files(self):
        """ Register all specified files and return a list of the final
        schema results
        """
        self.log.info(
            'Found the following *.{0} files: {1}'.format(
                self.file_extension,
                self.file_paths
            )
        )
        return [self.bootstrap_file(file_path) for file_path in self.file_paths]

    def bootstrap_file(self, file_path):
        """ Bootstrap a file with all metadata from schema_ref and
        return the final swaggerpy schema_result
        """
        schema_result = self.logged_register_file(file_path)
        return self.bootstrap_schema_result(schema_result)

    def bootstrap_schema_result(self, schema_result):
        """ Bootstraps a swaggerpy schema_result with all metadata from
        schema_ref (docs, notes, etc) and return the final schema_result.
        """
        source_ref = self.schema_ref.get_source_ref(
            schema_result.topic.source.name
        )
        if not source_ref:
            return schema_result

        # Updating the docs first, since it requires re-registering with
        # a modified avro schema
        schema_result = self.register_schema_docs(
            schema_result=schema_result,
            source_ref=source_ref
        )

        # Now we can register other attributes which have their own API
        # endpoints, such as categories, file_sources, and notes.
        self.register_schema_note(
            schema_result=schema_result,
            note=self.schema_ref.get_ref_val(source_ref, 'note')
        )
        self.register_category(
            schema_result=schema_result,
            category=self.schema_ref.get_ref_val(source_ref, 'category')
        )
        self.register_file_source(
            schema_result=schema_result,
            display=self.schema_ref.get_ref_val(source_ref, 'file_display'),
            url=self.schema_ref.get_ref_val(source_ref, 'file_url')
        )
        self.register_fields_notes(
            schema_result=schema_result,
            schema_json=json.loads(schema_result.schema),
            fields_ref=source_ref.get('fields', [])
        )
        return schema_result

    def logged_api_call(self, api_method, **kwargs):
        self.log.info(" - Calling {} with args:\n{}".format(
            repr(api_method),
            pformat(kwargs)
        ))
        result = api_method(**kwargs).result()
        self.log.info(" - Result: {}".format(result))
        return result

    def register_schema_docs(self, schema_result, source_ref):
        """ Update the "doc" attributes of the schema and all it's fields,
        returning back the result of the register_schema call.
        """
        source = schema_result.topic.source.name
        schema_json = json.loads(schema_result.schema)
        if self.override_metadata or not schema_json.get('doc'):
            schema_json['doc'] = self.schema_ref.get_ref_val(source_ref, 'doc')
        schema_json = self.update_field_docs(
            schema_json=schema_json,
            fields_ref=source_ref.get('fields', [])
        )
        return self.logged_api_call(
            self.api.schemas.register_schema,
            body={
                'base_schema_id': schema_result.schema_id,
                'schema': json.dumps(schema_json, cls=FrozenDictEncoder),
                'namespace': self.schema_ref.get_source_val(
                    source,
                    'namespace'
                ),
                'source': source,
                'source_owner_email': self.schema_ref.get_source_val(
                    source,
                    'owner_email'
                ),
                'contains_pii': self.schema_ref.get_source_val(
                    source,
                    'contains_pii'
                )
            }
        )

    def update_field_docs(self, schema_json, fields_ref):
        """ Update the "doc" attributes of the fields of a schema.
        """
        field_to_ref_map = {
            field_ref['name']: field_ref for field_ref in fields_ref
        }
        for field_json in schema_json['fields']:
            field_ref = field_to_ref_map.get(field_json['name'])
            if self.override_metadata or not field_json.get('doc'):
                field_json['doc'] = self.schema_ref.get_ref_val(
                    field_ref,
                    'doc'
                )
        return schema_json

    def register_schema_note(self, schema_result, note):
        if note is None:
            return
        if schema_result.note:
            if self.override_metadata:
                self.logged_api_call(
                    self.api.notes.update_note,
                    note_id=schema_result.note.id,
                    body={
                        'note': note,
                        'last_updated_by': self.schema_ref.doc_owner
                    }
                )
        else:
            self.logged_api_call(
                self.api.notes.create_note,
                body={
                    'reference_id': schema_result.schema_id,
                    'reference_type': 'schema',
                    'note': note,
                    'last_updated_by': self.schema_ref.doc_owner
                }
            )

    def register_category(self, schema_result, category):
        if category is None:
            return
        source_result = schema_result.topic.source

        if self.override_metadata or not source_result.category:
            self.logged_api_call(
                self.api.sources.update_category,
                source_id=source_result.source_id,
                body={'category': category}
            )

    def register_file_source(self, schema_result, display, url):
        # TODO (joshszep|DATAPIPE-324) - after file source API is implemented
        pass

    def register_fields_notes(self, schema_result, schema_json, fields_ref):
        field_to_ref_map = {
            field_ref['name']: field_ref for field_ref in fields_ref
        }
        results = self.logged_api_call(
            self.api.schemas.get_schema_elements_by_schema_id,
            schema_id=schema_result.schema_id
        )
        field_to_schema_element_map = {
            self.get_field_name_from_schema_element(result): result
            for result in results
            if result.element_type != 'record'
        }
        for field_json in schema_json['fields']:
            field_name = field_json['name']
            field_ref = field_to_ref_map.get(field_name)
            if not field_ref:
                continue
            self.register_schema_element_note(
                schema_element=field_to_schema_element_map[field_name],
                note=self.schema_ref.get_ref_val(field_ref, 'note')
            )

    def get_field_name_from_schema_element(self, schema_element):
        return schema_element.key.split('|')[-1]

    def register_schema_element_note(self, schema_element, note):
        if note is None:
            return
        if schema_element.note:
            if self.override_metadata:
                return self.logged_api_call(
                    self.api.notes.update_note,
                    note_id=schema_element.note.id,
                    body={
                        'note': note,
                        'last_updated_by': self.schema_ref.doc_owner
                    }
                )
        else:
            return self.logged_api_call(
                self.api.notes.create_note,
                body={
                    'reference_id': schema_element.id,
                    'reference_type': 'schema_element',
                    'note': note,
                    'last_updated_by': self.schema_ref.doc_owner
                }
            )


class AVSCBootstrapper(FileBootstrapperBase):
    """ A file bootstrapper to bootstrap Avro *.avsc json files. Expects all
    files to be formatted following the specification for a data pipeline
    message. See `data_pipeline/schemas/monitoring_message_v1.avsc` for
    an example.
    """

    def __init__(self, schema_ref, file_paths, override_metadata):
        super(AVSCBootstrapper, self).__init__(
            schema_ref, file_paths, override_metadata, file_extension='avsc'
        )

    def register_file(self, file_path):
        with open(file_path) as avsc_file:
            avsc_content = avsc_file.read()
            return self.register_avsc(avsc_content)

    def register_avsc(self, avsc_content):
        self.log.info(
            '\n - Registering avsc content:\n{}'.format(avsc_content)
        )
        avsc = json.loads(avsc_content)
        source = avsc['name']
        return self.logged_api_call(
            self.api.schemas.register_schema,
            body={
                'schema': avsc_content,
                'namespace': avsc['namespace'],
                'source': source,
                'source_owner_email': self.schema_ref.get_source_val(
                    source,
                    'owner_email'
                ),
                'contains_pii': self.schema_ref.get_source_val(
                    source,
                    'contains_pii'
                )
            }
        )


class MySQLBootstrapper(FileBootstrapperBase):
    """ A file bootstrapper to bootstrap MySQL *.sql files. Expects all
    files to contain a CREATE TABLE statement and nothing more.
    """

    def __init__(self, schema_ref, file_paths, override_metadata):
        super(MySQLBootstrapper, self).__init__(
            schema_ref, file_paths, override_metadata, file_extension='sql'
        )

    def register_file(self, file_path):
        with open(file_path) as sql_file:
            return self.register_sql(
                sql_content=sql_file.read(),
                source=self.get_source_from_sql_file_path(file_path)
            )

    def register_sql(self, sql_content, source):
        self.log.info(
            '\n - Registering sql content:\n{}'.format(sql_content)
        )
        return self.logged_api_call(
            self.api.schemas.register_schema_from_mysql_stmts,
            body={
                'new_create_table_stmt': sql_content,
                'namespace': self.schema_ref.get_source_val(
                    source,
                    'namespace'
                ),
                'source': source,
                'source_owner_email': self.schema_ref.get_source_val(
                    source,
                    'owner_email'
                ),
                'contains_pii': self.schema_ref.get_source_val(
                    source,
                    'contains_pii'
                )
            }
        )

    def get_source_from_sql_file_path(self, sql_file_path):
        """ Get a source name for the source by chopping off the `.sql` from
        the filename. Luckily all our tables follow this naming convention.
        """
        return path.split(sql_file_path)[1][:-4]


def is_file_bootstrapper_class(obj):
    return inspect.isclass(obj) and FileBootstrapperBase in obj.__bases__


FILE_BOOTSTRAPPER_CLASSES = [
    obj for _, obj in inspect.getmembers(sys.modules[__name__], is_file_bootstrapper_class)
]


class BootstrapperBatch(yelp_batch.batch.Batch):

    notify_emails = ['bam+batch@yelp.com']

    @batch_command_line_options
    def parse_options(self, option_parser):
        opt_group = optparse.OptionGroup(option_parser, "Bootstrapper Options")
        opt_group.add_option(
            '--glob',
            action='append',
            type='string',
            default=[],
            dest='globs',
            help='[REQUIRED] Either a path to a specific MySQL *.sql file, a '
                 'specific Avro *.avsc file, or a glob pattern for a directory '
                 'containing such files. (For example: '
                 '"/nail/home/USER/pg/yelp-main/schema/yelp_dw/tables/*.sql") '
                 'Note --glob may be provided multiple times.'
        )
        opt_group.add_option(
            '--schema-ref',
            type='string',
            help='File path to load the schema reference json document from.'
                 'See DATAPIPE-259 for more'
                 'information on the schema reference json format.'
        )
        opt_group.add_option(
            '--default-source-owner',
            default='bam@yelp.com',
            type='string',
            help='Owner email of the bootstrapped tables, if none provided by '
                 'the schema-ref. '
                 'Default is %default'
        )
        opt_group.add_option(
            '--default-doc-owner',
            default='bam@yelp.com',
            type='string',
            help='Email of the default owner for the documentation of '
                 'bootstrapped tables, if none is provided by the schema-ref. '
                 'The documentation owner will be known as the last to update '
                 'the descriptions and notes within the bootstrapped tables. '
                 'Default is %default'
        )
        opt_group.add_option(
            '--default-namespace',
            type='string',
            help='Default namespace to register schemas to, if none provided '
                 'by the schema-ref.'
        )
        opt_group.add_option(
            '--default-docstring',
            default='',
            type='string',
            help='Default documentation string for any table descriptions and'
                 'column descriptions which are not provided by the'
                 'schema-ref. '
                 'Default is "%default"'
        )
        opt_group.add_option(
            '--default-contains-pii',
            action="store_true",
            default=False,
            help='Default to mark schemas as containing pii, if not otherwise'
                 'specified by the schema-ref. '
                 'Default is "%default"'
        )
        opt_group.add_option(
            '--default-category',
            type='string',
            default=None,
            help='Default category for schemas, if none provided by the '
                 'schema-ref. If this is not provided the schemas will be '
                 '"[ Uncategorized ]".'
        )
        opt_group.add_option(
            '--http-host',
            default='http://' + gethostname(),
            type='string',
            help='Web hostname. Only used for reporting the links to the doc'
                 'tool view of the schemas which were registered during the '
                 'run. '
                 'Default is "%default"'
        )
        opt_group.add_option(
            '--override-metadata',
            action="store_true",
            default=False,
            help='Override existing metadata for tables (such as notes, '
                 'categories, etc) with provided metadata. By default if'
                 'metadata already exists it will not be changed.'
                 'Default is "%default"'
        )
        return opt_group

    def run(self):
        """ Primary entry point for the batch
        """

        schema_ref = SchemaRef.load_from_file(
            schema_ref_path=self.options.schema_ref,
            defaults={
                'doc_owner': self.options.default_doc_owner,
                'owner_email': self.options.default_source_owner,
                'namespace': self.options.default_namespace,
                'doc': self.options.default_docstring,
                'contains_pii': bool(self.options.default_contains_pii),
                'category': self.options.default_category
            }
        )
        FileBootstrapperBase.log = self.log  # Specify our logger as the logger
        file_paths = get_file_paths_from_glob_patterns(
            self.options.globs
        )
        schema_results = []
        for bootstrapper_cls in FILE_BOOTSTRAPPER_CLASSES:
            bootstrapper = bootstrapper_cls(
                schema_ref=schema_ref,
                file_paths=file_paths,
                override_metadata=self.options.override_metadata
            )
            schema_results += bootstrapper.bootstrap_files()

        self.log_result_urls(schema_results)

    def get_files_from_glob_patterns(self, glob_patterns):
        """ Return a list of files matching the given list of glob patterns
         (for example ["./test.sql", "./other_tables/*.sql"])
        """
        file_paths = []
        for glob_pattern in glob_patterns:
            file_paths += glob.glob(glob_pattern)
        return file_paths

    def log_result_urls(self, schema_results):
        self.log.info("Completed updating the following tables:")
        for schema_result in schema_results:
            self.log.info(
                '{host}/web/#/table?schema={namespace}&table={source}'.format(
                    host='{}:{}'.format(
                        self.options.http_host,
                        get_config().schematizer_port
                    ),
                    namespace=schema_result.topic.source.namespace.name,
                    source=schema_result.topic.source.name
                )
            )


if __name__ == "__main__":
    BootstrapperBatch().start()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel

"""
Represent the data of a note. A note is used to document the data.
For example, the text used to explain the meaning of the is_inactive column.

Args:
    id (int): the id of the note.
    reference id (int): the id of the entity that this note describes
    reference_type (str): indicate the type of the entity (schema or schema_elements)
    updated_at (timestamp): the updated time
    note (str): the text of this note
    last_updated_by (str): the entity that last updated this note
    created_at (timestamp): the time that this note was created
"""

Note = namedtuple(
    'Note',
    [
        'created_at',
        'reference_type',
        'updated_at',
        'note',
        'last_updated_by',
        'reference_id',
        'id'
    ]
)


class _Note(BaseModel):
    """Internal class used to convert from/to various data structures and
    facilitate constructing the return value of schematizer functions.
    """

    def __init__(
        self,
        id,
        reference_id,
        reference_type,
        created_at,
        updated_at,
        note,
        last_updated_by
    ):
        self.id = id
        self.reference_id = reference_id
        self.reference_type = reference_type
        self.created_at = created_at
        self.updated_at = updated_at
        self.note = note
        self.last_updated_by = last_updated_by

    @classmethod
    def from_response(cls, response):
        if response is None:
            return None
        return cls(
            id=response.id,
            reference_id=response.reference_id,
            reference_type=response.reference_type,
            created_at=response.created_at,
            updated_at=response.updated_at,
            note=response.note,
            last_updated_by=response.last_updated_by
        )

    def to_cache_value(self):
        return {
            'id': self.id,
            'reference_id': self.reference_id,
            'reference_type': self.reference_type,
            'created_at': self.created_at,
            'updated_at': self.updated_at,
            'note': self.note,
            'last_updated_by': self.last_updated_by
        }

    @classmethod
    def from_cache_value(cls, cache_value):
        if cache_value is None:
            return None
        return cls(
            id=cache_value['id'],
            reference_id=cache_value['reference_id'],
            reference_type=cache_value['reference_type'],
            created_at=cache_value['created_at'],
            updated_at=cache_value['updated_at'],
            note=cache_value['note'],
            last_updated_by=cache_value['last_updated_by']
        )

    def to_result(self):
        return Note(
            id=self.id,
            reference_id=self.reference_id,
            reference_type=self.reference_type,
            created_at=self.created_at,
            updated_at=self.updated_at,
            note=self.note,
            last_updated_by=self.last_updated_by
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from datetime import datetime
from datetime import timedelta

from dateutil.tz import tzutc


class BasePeriodicProcessor(object):
    """ This class provides an interface for handling periodic events that can
    be triggered by a heartbeat event, like sensu alert and data event checkpoint.
    That is, it serves as a base class and must be subclassed.

    Args:
      interval_in_seconds(int): the time interval between two events.
    """

    def __init__(self, interval_in_seconds):
        self.interval_in_seconds = interval_in_seconds
        self._next_process_time = self._utc_now

    def periodic_process(self, timestamp=None):
        """ This method remains because it's called by the replication handler;
        if / when we start calling the process method below directly from the
        replication handler we can remove it (DATAPIPE-1435)
        Args:
            timestamp(datetime.datetime): the datetime of the event with utc
        """
        if self._should_process():
            self.process(timestamp)
            self._next_process_time = self._compute_next_process_time()

    def process(self, timestamp=None):
        raise NotImplementedError

    def _should_process(self):
        return self._utc_now >= self._next_process_time

    def _compute_next_process_time(self):
        return self._utc_now + timedelta(seconds=self.interval_in_seconds)

    @property
    def _utc_now(self):
        return datetime.now(tzutc())



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import os

from Crypto.Cipher import AES

from data_pipeline.meta_attribute import MetaAttribute


def get_initialization_vector(schema_id, initialization_vector_array=None):
    if initialization_vector_array is None:
        initialization_vector_array = os.urandom(AES.block_size)
    _verify_initialization_vector_params(initialization_vector_array)
    return MetaAttribute(
        schema_id=schema_id,
        payload_data=initialization_vector_array
    )


def _verify_initialization_vector_params(vector_array):
    if not isinstance(vector_array, bytes) or not len(vector_array) == 16:
        raise TypeError('Initialization Vector must be a 16-byte array')



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import simplejson
from bravado.exception import HTTPNotFound
from requests.exceptions import RequestException
from swagger_zipkin.zipkin_decorator import ZipkinClientDecorator

from data_pipeline._retry_util import ExpBackoffPolicy
from data_pipeline._retry_util import retry_on_exception
from data_pipeline._retry_util import RetryPolicy
from data_pipeline.config import get_config
from data_pipeline.helpers.singleton import Singleton
from data_pipeline.schematizer_clientlib.models.avro_schema import _AvroSchema
from data_pipeline.schematizer_clientlib.models.avro_schema_element import (
    _AvroSchemaElement
)
from data_pipeline.schematizer_clientlib.models.consumer_group import (
    _ConsumerGroup
)
from data_pipeline.schematizer_clientlib.models.consumer_group_data_source import (
    _ConsumerGroupDataSource
)
from data_pipeline.schematizer_clientlib.models.data_target import _DataTarget
from data_pipeline.schematizer_clientlib.models.meta_attr_namespace_mapping import (
    _MetaAttributeNamespaceMapping
)
from data_pipeline.schematizer_clientlib.models.meta_attr_source_mapping import (
    _MetaAttributeSourceMapping
)
from data_pipeline.schematizer_clientlib.models.namespace import _Namespace
from data_pipeline.schematizer_clientlib.models.refresh import _Refresh
from data_pipeline.schematizer_clientlib.models.source import _Source
from data_pipeline.schematizer_clientlib.models.topic import _Topic


class _Cache(object):
    """Cache used by Schematizer client.  This cache stores the schematizer
    entities, such as avro schemas, topics, sources, etc.

    This cache is currently limited to used by SchematizerClient only.  The
    cached value is expected to have `to_cache_value` and `from_cache_value`
    functions.  This limitation could be relaxed later if necessary.
    """

    def __init__(self):
        self._cache = {}

    def get_value(self, entity_type, entity_key):
        cache_key = self._get_cache_key(entity_type.__name__, entity_key)
        cache_value = self._cache.get(cache_key)
        return entity_type.from_cache_value(cache_value) if cache_value else None

    def set_value(self, entity_key, new_value):
        value_type_name = new_value.__class__.__name__
        cache_key = self._get_cache_key(value_type_name, entity_key)
        self._cache[cache_key] = new_value.to_cache_value()

    def _get_cache_key(self, entity_type_name, entity_key):
        return entity_type_name, entity_key


class SchematizerClient(object):
    """A client that interacts with Schematizer APIs.  It has built-in caching
    feature which caches avro schemas, topics, and etc.  Right now the cache is
    only in memory (TODO(DATAPIPE-162|joshszep): Implement persistent caching).

    It caches schemas, topics, and sources separately instead of caching nested
    objects to avoid storing duplicate data repeatedly.

    Currently the client will throw the HTTPError returned from the Schematizer
    service if an error occurs.  It could be nice to return more straight forward
    errors instead of HTTPError.  The work is tracked in DATAPIPE-352.
    """

    # This class potentially could grow relatively huge.  There may be a need to
    # split them in some way later if it's too huge.  One idea is to split related
    # functions into its own mix-in class, optionally in its own file, and have main
    # `SchematizerClient` class inherit this mix-in class.  The benefit of this
    # approach is the size of each mix-in class is manageable and easier to read.
    # The down side is it becomes harder to know if a function has been defined
    # since they're now in multiple classes/files.

    # TODO[clin|DATAPIPE-1518] change the rest of the hidden variables to normal
    # variables.

    __metaclass__ = Singleton

    # Default page size used for the pagination when calling bulk apis such as
    # `get_topics_by_criteria`. If a bulk api needs pagination, this should be
    # the page size by default. If the api needs custom page size, we should
    # consider a good way to implement it because I'd like to avoid adding a
    # bunch of such constants which clutter the code.
    DEFAULT_PAGE_SIZE = 20

    def __init__(self):
        self._bravado_client = get_config().schematizer_client
        self._client = ZipkinClientDecorator(self._bravado_client)
        self._cache = _Cache()

    def get_schema_by_id(self, schema_id):
        """Get the avro schema of given schema id.

        Args:
            schema_id (int): The id of requested avro schema.

        Returns:
            (data_pipeline.schematizer_clientlib.models.avro_schema.AvroSchema):
                The requested avro Schema.
        """
        return self._get_schema_by_id(schema_id).to_result()

    def _get_schema_by_id(self, schema_id):
        _schema = self._get_cached_schema(schema_id)
        if _schema:
            return _schema

        response = self._call_api(
            api=self._client.schemas.get_schema_by_id,
            params={'schema_id': schema_id}
        )
        _schema = _AvroSchema.from_response(response)
        self._set_cache_by_schema(_schema)
        return _schema

    def get_schema_elements_by_schema_id(self, schema_id):
        """Get the avro schema elements of given schema id.

        Args:
            schema_id (int): The id of requested avro schema elements.

        Returns:
            (List of data_pipeline.schematizer_clientlib.models.avro_schema_element.AvroSchemaElement):
                The list requested avro Schema elements by schema_id.
        """
        # Filter out elements that represent the whole record (when element.element_name == None)
        return [
            element.to_result()
            for element in self._get_schema_elements_by_schema_id(schema_id)
            if element.element_name
        ]

    def _get_schema_elements_by_schema_id(self, schema_id):
        response = self._call_api(
            api=self._client.schemas.get_schema_elements_by_schema_id,
            params={'schema_id': schema_id}
        )
        _schema = _AvroSchemaElement.from_response(response)
        return _schema

    def get_schemas_created_after_date(
        self,
        created_after,
        min_id=0,
        page_size=10
    ):
        """ Get the avro schemas (excluding disabled schemas) created after the
        given datetime timestamp in ascending order of schema id. Limits the
        result to those with id greater than or equal to the min_id.

        Args:
            created_after (long): get schemas created at or after the given
                epoch timestamp.
            min_id (Optional[int]): Limits the result to those schemas with an
                id greater than or equal to given min_id (default: 0)
            page_size (Optional[int]): Limits the number of api calls and
                number of schemas to retrieve per call to avoid timeouts
                (default: 10).

        Returns:
            (List of data_pipeline.schematizer_clientlib.models.avro_schema.AvroSchema):
                The list of avro schemas created after (inclusive) specified date.
        """
        return self._get_schemas_created_after_date(
            created_after,
            min_id,
            page_size
        )

    def get_schemas_by_criteria(
        self,
        created_after=0,
        min_id=0,
        count=10
    ):
        """ Get the avro schemas with the specified criteria: created_after,
        min_id and count.

        Args:
            created_after (Optional[long]): get schemas created at or after the given
                epoch timestamp (default 0)
            min_id (Optional[int]): Limits the result to those schemas with an
                id greater than or equal to given min_id (default: 0)
            count (Optional[int]): the maximum number of schemas returned in this call
                (default: 10).

        Returns:
            (List of data_pipeline.schematizer_clientlib.models.avro_schema.AvroSchema):
                list of avro_schemas satifying the criteria specified in parameters
        """
        results = []
        response = self._call_api(
            api=self._client.schemas.get_schemas_created_after,
            params={
                'created_after': created_after,
                'count': count,
                'min_id': min_id
            }
        )

        for resp_item in response:
            _schema = _AvroSchema.from_response(resp_item)
            results.append(_schema.to_result())
            self._set_cache_by_schema(_schema)
        return results

    def _get_schemas_created_after_date(self, created_after, min_id, page_size):
        last_page_size = page_size
        result = []
        # Exit the loop when the number of schemas returned are not equal to
        # page_size. Denoting there are no more schemas to fetch from the
        # Schematizer.
        while last_page_size == page_size:
            response = self._call_api(
                api=self._client.schemas.get_schemas_created_after,
                params={
                    'created_after': created_after,
                    'count': page_size,
                    'min_id': min_id
                }
            )

            for resp_item in response:
                _schema = _AvroSchema.from_response(resp_item)
                result.append(_schema.to_result())
                self._set_cache_by_schema(_schema)
                min_id = _schema.schema_id + 1
            last_page_size = len(response)
        return result

    def get_schemas_by_topic(self, topic_name):
        """Get the list of schemas in the specified topic.

        Args:
            topic_name (str): name of the topic to look up

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.avro_schema.AvroSchema):
                List of schemas in the given topic
        """
        response = self._call_api(
            api=self._client.topics.list_schemas_by_topic_name,
            params={"topic_name": topic_name}
        )
        result = []
        for resp_item in response:
            _schema = _AvroSchema.from_response(resp_item)
            result.append(_schema.to_result())
            self._set_cache_by_schema(_schema)
        return result

    def get_topic_by_name(self, topic_name):
        """Get the topic of given topic name.

        Args:
            topic_name (str): The name of requested topic.

        Returns:
            (data_pipeline.schematizer_clientlib.models.topic.Topic):
                The requested topic.
        """
        return self._get_topic_by_name(topic_name).to_result()

    def _get_topic_by_name(self, topic_name):
        _topic = self._get_cached_topic(topic_name)
        if _topic:
            return _topic

        response = self._call_api(
            api=self._client.topics.get_topic_by_topic_name,
            params={'topic_name': topic_name}
        )
        _topic = _Topic.from_response(response)
        self._set_cache_by_topic(_topic)
        return _topic

    def get_source_by_id(self, source_id):
        """Get the schema source of given source id.

        Args:
            source_id (int): The id of the source.

        Returns:
            (data_pipeline.schematizer_clientlib.models.topic.Source):
                The requested schema source.
        """
        return self._get_source_by_id(source_id).to_result()

    def _get_source_by_id(self, source_id):
        _source = self._cache.get_value(_Source, source_id)
        if _source:
            return _source

        response = self._call_api(
            api=self._client.sources.get_source_by_id,
            params={'source_id': source_id}
        )
        _source = _Source.from_response(response)
        self._set_cache_by_source(_source)
        return _source

    def get_namespaces(self):
        """Get the list of namespaces registered in the schematizer

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.namespace.Namespace])
                The list of namespaces
        """
        response = self._call_api(
            api=self._client.namespaces.list_namespaces,
            params={}
        )
        result = []
        for resp_item in response:
            _namespace = _Namespace.from_response(resp_item)
            result.append(_namespace.to_result())
        return result

    def get_sources_by_namespace(
        self,
        namespace_name,
        min_id=0,
        page_size=10
    ):
        """Get the list of sources in the specified namespace.

        Args:
            namespace_name (str): namespace name to look up
            min_id (Optional[int]): the returned sources should have id greater than or equal to the min_id
            page_size (Optional[int]): the number of sources to return in one api call to prevent timeout

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.source.Source]):
                The list of schema sources in the given namespace.
        """
        last_page_size = page_size
        result = []
        while last_page_size == page_size:
            response = self._call_api(
                api=self._client.namespaces.list_sources_by_namespace,
                params={
                    'namespace': namespace_name,
                    'min_id': min_id,
                    'count': page_size
                }
            )
            for resp_item in response:
                _source = _Source.from_response(resp_item)
                result.append(_source.to_result())
                self._set_cache_by_source(_source)
                min_id = _source.source_id + 1
            last_page_size = len(response)
        return result

    def get_sources(
        self,
        min_id=0,
        page_size=10
    ):
        """Get the sources that match specified criteria (min_id and
        page_size).  If no criterion is specified, it returns all the sources.

        Args:
            min_id (Optional[int]): Limits results to those sources with an id
                greater than or equal to given min_id (default: 0)
            page_size (Optional[int]): Maximum number of sources to retrieve
                per page. (default: 10)

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.Source]):
                list of topics that match given criteria.
        """
        last_page_size = page_size
        result = []
        while last_page_size == page_size:
            response = self._call_api(
                api=self._client.sources.list_sources,
                params={
                    'min_id': min_id,
                    'count': page_size
                }
            )
            for resp_item in response:
                _source = _Source.from_response(resp_item)
                result.append(_source.to_result())
                self._set_cache_by_source(_source)
                min_id = _source.source_id + 1
            last_page_size = len(response)
        return result

    def get_topics_by_source_id(self, source_id):
        """Get the list of topics of specified source id.

        Args:
            source_id (int): The id of the source to look up

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.topic.Topic]):
                The list of topics of given source.
        """
        response = self._call_api(
            api=self._client.sources.list_topics_by_source_id,
            params={'source_id': source_id}
        )
        result = []
        for resp_item in response:
            _topic = _Topic.from_response(resp_item)
            result.append(_topic.to_result())
            self._set_cache_by_topic(_topic)
        return result

    def get_latest_topic_by_source_id(self, source_id):
        """Get the lastest topic of specified source id

        Args:
            source_id (int): The id of the source to look up

        Returns:
            (data_pipeline.schematizer_clientlib.models.topic.Topic):
                The latest topic of given source
        """
        response = self._call_api(
            api=self._client.sources.get_latest_topic_by_source_id,
            params={'source_id': source_id}
        )
        _topic = _Topic.from_response(response)
        self._set_cache_by_topic(_topic)
        return _topic.to_result()

    def get_latest_schema_by_topic_name(self, topic_name):
        """Get the latest enabled schema of given topic.

        Args:
            topic_name (str): The name of the topic to look up

        Returns:
            (data_pipeline.schematizer_clientlib.models.avro_schema.AvroSchema):
                The latest enabled avro schema of given topic.  It returns None
                if no such avro schema can be found.
        """
        response = self._call_api(
            api=self._client.topics.get_latest_schema_by_topic_name,
            params={'topic_name': topic_name}
        )
        _schema = _AvroSchema.from_response(response)
        self._set_cache_by_schema(_schema)
        return _schema.to_result()

    def register_schema(
        self,
        namespace,
        source,
        schema_str,
        source_owner_email,
        contains_pii,
        cluster_type='datapipe',
        base_schema_id=None
    ):
        """ Register a new schema and return newly created schema object.

        Args:
            namespace (str): The namespace the new schema is registered to.
            source (str): The source the new schema is registered to.
            schema_str (str): String representation of the avro schema.
            source_owner_email (str): The owner email of the given source.
            contains_pii (bool): Indicates if the schema being registered has
                at least one field that can potentially contain PII.
                See http://y/pii for help identifying what is or is not PII.
            cluster_type (Optional[str]): Kafka cluster type to connect to,
                like 'datapipe', 'scribe', etc. See http://y/datapipe_cluster_types
                for more info on cluster_types. Defaults to datapipe.
            base_schema_id (Optional[int]): The id of the original schema which
                the new schema was changed based on

        Returns:
            (data_pipeline.schematizer_clientlib.models.avro_schema.AvroSchema):
                The newly created avro Schema.
        """
        request_body = {
            'schema': schema_str,
            'namespace': namespace,
            'source': source,
            'source_owner_email': source_owner_email,
            'contains_pii': contains_pii,
            'cluster_type': cluster_type
        }
        if base_schema_id:
            request_body['base_schema_id'] = base_schema_id
        response = self._call_api(
            api=self._client.schemas.register_schema,
            request_body=request_body
        )

        _schema = _AvroSchema.from_response(response)
        self._set_cache_by_schema(_schema)
        return _schema.to_result()

    def register_namespace_meta_attribute_mapping(
        self,
        namespace_name,
        meta_attr_schema_id
    ):
        """ Register a meta attribute for the given namespace.

        Args:
            namespace_name (str): The name of the namespace.
            meta_attr_schema_id (int): ID of the meta attribute (avro_schema)

        Returns:
            (data_pipeline.schematizer_clientlib.models.
            meta_attr_namespace_mapping.MetaAttributeNamespaceMapping): The
            newly created mapping between specified namespace and
            schema ID.
        """
        response = self._call_api(
            api=self._client.namespaces.register_namespace_meta_attribute_mapping,
            params={'namespace': namespace_name},
            request_body={
                'meta_attribute_schema_id': meta_attr_schema_id
            }
        )
        _meta_attr_mapping = _MetaAttributeNamespaceMapping.from_response(
            namespace_id=response.namespace_id,
            meta_attribute_schema_id=response.meta_attribute_schema_id
        )
        return _meta_attr_mapping.to_result()

    def delete_namespace_meta_attribute_mapping(
        self,
        namespace_name,
        meta_attr_schema_id
    ):
        """ Deletes a meta attribute from the given namespace.

        Args:
            namespace_name (str): The name of the namespace.
            meta_attr_schema_id (int): ID of the meta attribute (avro_schema)

        Returns:
            (data_pipeline.schematizer_clientlib.models.
            meta_attr_namespace_mapping.MetaAttributeNamespaceMapping): A
            successfully deleted mapping between specified namespace and
            schema ID.
        """
        response = self._call_api(
            api=self._client.namespaces.delete_namespace_meta_attribute_mapping,
            params={'namespace': namespace_name},
            request_body={
                'meta_attribute_schema_id': meta_attr_schema_id
            }
        )
        _meta_attr_mapping = _MetaAttributeNamespaceMapping.from_response(
            namespace_id=response.namespace_id,
            meta_attribute_schema_id=response.meta_attribute_schema_id
        )
        return _meta_attr_mapping.to_result()

    def get_namespace_meta_attribute_mappings(self, namespace_name):
        """ Returns meta attributes for the given namespace.

        Args:
            namespace_name (str): The name of the namespace.

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.
            meta_attr_namespace_mapping.MetaAttributeNamespaceMapping]): A
            list of mappings between specified namespace and schema IDs.
        """
        response = self._call_api(
            api=self._client.namespaces.get_namespace_meta_attribute_mappings,
            params={'namespace': namespace_name}
        )
        return [
            _MetaAttributeNamespaceMapping.from_response(
                namespace_id=resp_item.namespace_id,
                meta_attribute_schema_id=resp_item.meta_attribute_schema_id
            ).to_result()
            for resp_item in response]

    def register_source_meta_attribute_mapping(
        self,
        source_id,
        meta_attr_schema_id
    ):
        """ Register a meta attribute for the given source.

        Args:
            source_id (int): The ID of the source.
            meta_attr_schema_id (int): ID of the meta attribute (avro_schema)

        Returns:
            (data_pipeline.schematizer_clientlib.models.
            meta_attr_source_mapping.MetaAttributeSourceMapping): The
            newly created mapping between specified source and
            schema ID.
        """
        response = self._call_api(
            api=self._client.sources.register_source_meta_attribute_mapping,
            params={'source_id': source_id},
            request_body={
                'meta_attribute_schema_id': meta_attr_schema_id
            }
        )
        _meta_attr_mapping = _MetaAttributeSourceMapping.from_response(
            source_id=response.source_id,
            meta_attribute_schema_id=response.meta_attribute_schema_id
        )
        return _meta_attr_mapping.to_result()

    def delete_source_meta_attribute_mapping(
        self,
        source_id,
        meta_attr_schema_id
    ):
        """ Deletes a meta attribute from the given source.

        Args:
            source_id (int): The ID of the source.
            meta_attr_schema_id (int): ID of the meta attribute (avro_schema)

        Returns:
            (data_pipeline.schematizer_clientlib.models.
            meta_attr_source_mapping.MetaAttributeSourceMapping): A
            successfully deleted mapping between the specified source and
            schema.
        """
        response = self._call_api(
            api=self._client.sources.delete_source_meta_attribute_mapping,
            params={'source_id': source_id},
            request_body={
                'meta_attribute_schema_id': meta_attr_schema_id
            }
        )
        _meta_attr_mapping = _MetaAttributeSourceMapping.from_response(
            source_id=response.source_id,
            meta_attribute_schema_id=response.meta_attribute_schema_id
        )
        return _meta_attr_mapping.to_result()

    def get_source_meta_attribute_mappings(self, source_id):
        """ Returns meta attributes for the given source.

        Args:
            source_id (int): The ID of the source.

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.
            meta_attr_source_mapping.MetaAttributeSourceMapping]): A
            list of mappings between the specified source and schema IDs.
        """
        response = self._call_api(
            api=self._client.sources.get_source_meta_attribute_mappings,
            params={'source_id': source_id}
        )
        result = []
        for resp_item in response:
            _meta_attr_mapping = _MetaAttributeSourceMapping.from_response(
                source_id=resp_item.source_id,
                meta_attribute_schema_id=resp_item.meta_attribute_schema_id
            )
            result.append(_meta_attr_mapping.to_result())
        return result

    def get_meta_attributes_by_schema_id(self, schema_id):
        """ Returns meta attributes for the given schema.

        Args:
            schema_id (int): The ID of the Avro Schema.

        Returns:
            (List[int]): A list of meta attribute ids for the given Avro
            Schema.
        """
        return self._call_api(
            api=self._client.schemas.get_meta_attributes_by_schema_id,
            params={'schema_id': schema_id}
        )

    def register_schema_from_schema_json(
        self,
        namespace,
        source,
        schema_json,
        source_owner_email,
        contains_pii,
        cluster_type='datapipe',
        base_schema_id=None
    ):
        """ Register a new schema and return newly created schema object.

        Args:
            namespace (str): The namespace the new schema is registered to.
            source (str): The source the new schema is registered to.
            schema_json (dict or list): Python object representation of the
                avro schema json.
            source_owner_email (str): The owner email of the given source.
            contains_pii (bool): Indicates if the schema being registered has
                at least one field that can potentially contain PII.
                See http://y/pii for help identifying what is or is not PII.
            cluster_type (Optional[str]): Kafka cluster type to connect like
                datapipe, scribe, etc. See http://y/datapipe_cluster_types for
                more info on cluster_types. Defaults to datapipe.
            base_schema_id (Optional[int]): The id of the original schema which
                the new schema was changed based on

        Returns:
            (data_pipeline.schematizer_clientlib.models.avro_schema.AvroSchema):
                The newly created avro Schema.
        """
        return self.register_schema(
            namespace=namespace,
            source=source,
            schema_str=simplejson.dumps(schema_json),
            source_owner_email=source_owner_email,
            contains_pii=contains_pii,
            cluster_type=cluster_type,
            base_schema_id=base_schema_id
        )

    def register_schema_from_mysql_stmts(
        self,
        namespace,
        source,
        source_owner_email,
        contains_pii,
        new_create_table_stmt,
        old_create_table_stmt=None,
        alter_table_stmt=None,
    ):
        """ Register schema based on mysql statements and return newly created
        schema.

        Args:
            namespace (str): The namespace the new schema is registered to.
            source (str): The source the new schema is registered to.
            source_owner_email (str): The owner email of the given source.
            contains_pii (bool): The flag indicating if schema contains pii.
            new_create_table_stmt (str): the mysql statement of creating new table.
            old_create_table_stmt (Optional[str]): the mysql statement of
                creating old table.
            alter_table_stmt (Optional[str]): the mysql statement of altering
                table schema.

        Returns:
            (data_pipeline.schematizer_clientlib.models.avro_schema.AvroSchema):
                The newly created avro Schema.
        """
        request_body = {
            'namespace': namespace,
            'source': source,
            'new_create_table_stmt': new_create_table_stmt,
            'source_owner_email': source_owner_email,
            'contains_pii': contains_pii
        }
        if old_create_table_stmt:
            request_body['old_create_table_stmt'] = old_create_table_stmt
        if alter_table_stmt:
            request_body['alter_table_stmt'] = alter_table_stmt
        response = self._call_api(
            api=self._client.schemas.register_schema_from_mysql_stmts,
            request_body=request_body
        )

        _schema = _AvroSchema.from_response(response)
        self._set_cache_by_schema(_schema)
        return _schema.to_result()

    def get_topics_by_criteria(
        self,
        namespace_name=None,
        source_name=None,
        created_after=None,
        min_id=0,
        max_count=None
    ):
        """Get all the topics that match specified criteria.  If no criterion
        is specified, it returns all the topics.

        Args:
            namespace_name (Optional[str]): namespace the topics belong to
            source_name (Optional[str]): name of the source topics belong to
            created_after (Optional[int]): Epoch timestamp the topics should be
                created after.  The topics created at the same timestamp are
                also included.
            min_id (Optional[int]): Limits results to those topics with an id
                greater than or equal to given min_id (default: 0)
            max_count (Optional[int]): Maximum number of topics to retrieve. It
                must be a positive integer. If not specified, it returns all the
                topics.

        Returns:
            List[data_pipeline.schematizer_clientlib.models.topic.Topic]:
                list of topics that match given criteria. The returned topics
                are ordered by their topic id.

        Remarks:
            The function internally paginates through the topics if the max_count
            is too large to avoid timeout from the service.  The page size is set
            to 20 right now.
        """
        # The reason to set fixed page size internally is to make the pagination
        # transparent to the users. We can set a reasonable page size that makes
        # reasonable number of calls without causing timeout. This interface
        # should be sufficient for most of our use cases, so we probably don't
        # need to provide the interface for users to set the page size right now.
        max_count_specified = max_count is not None and max_count > 0
        page_size = self.DEFAULT_PAGE_SIZE
        should_get_more_topics = True
        result = []
        while should_get_more_topics:
            response = self._call_api(
                api=self._client.topics.get_topics_by_criteria,
                params={
                    'namespace': namespace_name,
                    'source': source_name,
                    'created_after': created_after,
                    'min_id': min_id,
                    'count': page_size
                }
            )
            topic = None
            for resp_item in response:
                topic = _Topic.from_response(resp_item)
                result.append(topic.to_result())
                self._set_cache_by_topic(topic)
                if max_count_specified and len(result) == max_count:
                    should_get_more_topics = False
                    break
            if topic:
                min_id = topic.topic_id + 1
            should_get_more_topics = (
                should_get_more_topics and len(response) >= page_size
            )
        return result

    def create_data_target(self, name, target_type, destination):
        """ Create and return newly created data target.

        Args:
            name (str): Name to uniquely identify the data target.
            target_type (str): The type of the data target, such as Redshift.
            destination (str): The actual location of the data target, such as
                Url of the Redshift cluster.

        Returns:
            (data_pipeline.schematizer_clientlib.models.data_target.DataTarget):
                The newly created data target.
        """
        response = self._call_api(
            api=self._client.data_targets.create_data_target,
            request_body={
                'name': name,
                'target_type': target_type,
                'destination': destination
            }
        )
        _data_target = _DataTarget.from_response(response)
        self._set_cache_by_data_target(_data_target)
        return _data_target.to_result()

    def get_data_targets_by_schema_id(self, schema_id):
        """Get data targets of specified schema id

        Args:
            schema_id (int): the id of a schema

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.data_target.DataTarget]):
                The list of data targets corresponding to the schema_id
        """
        return self._get_data_targets_by_schema_id(schema_id)

    def _get_data_targets_by_schema_id(self, schema_id):
        response = self._call_api(
            api=self._client.schemas.get_data_targets_by_schema_id,
            params={'schema_id': schema_id}
        )
        results = []
        for resp_item in response:
            _data_target = _DataTarget.from_response(resp_item)
            results.append(_data_target)
            self._set_cache_by_data_target(_data_target)
        return results

    def get_data_target_by_id(self, data_target_id):
        """Get the data target of specified id.

        Args:
            data_target_id (int): The id of requested data target.

        Returns:
            (data_pipeline.schematizer_clientlib.models.data_target.DataTarget):
                The requested data target.
        """
        return self._get_data_target_by_id(data_target_id).to_result()

    def _get_data_target_by_id(self, data_target_id):
        _data_target = self._cache.get_value(_DataTarget, data_target_id)
        if _data_target:
            return _data_target

        response = self._call_api(
            api=self._client.data_targets.get_data_target_by_id,
            params={'data_target_id': data_target_id}
        )
        _data_target = _DataTarget.from_response(response)
        self._set_cache_by_data_target(_data_target)
        return _data_target

    def get_data_target_by_name(self, data_target_name):
        """Get the data target of specified name.

        Args:
            data_target_name (str): The name of requested data target.

        Returns:
            (data_pipeline.schematizer_clientlib.models.data_target.DataTarget):
                The requested data target.
        """
        return self._get_data_target_by_name(data_target_name).to_result()

    def _get_data_target_by_name(self, data_target_name):
        _data_target = self._cache.get_value(_DataTarget, data_target_name)
        if _data_target:
            return _data_target

        response = self._call_api(
            api=self._client.data_targets.get_data_target_by_name,
            params={'data_target_name': data_target_name}
        )
        _data_target = _DataTarget.from_response(response)
        self._set_cache_by_data_target_name(_data_target)
        return _data_target

    def get_topics_by_data_target_id(self, data_target_id):
        """Get the list of topics associated to the specified data target id.

        Args:
            data_target_id (int): The id of the data target to look up

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.topic.Topic]):
                The list of topics of given data target.
        """
        response = self._call_api(
            api=self._client.data_targets.get_topics_by_data_target_id,
            params={'data_target_id': data_target_id}
        )
        result = []
        for resp_item in response:
            _topic = _Topic.from_response(resp_item)
            result.append(_topic.to_result())
            self._set_cache_by_topic(_topic)
        return result

    def get_refreshes_by_criteria(
        self,
        namespace_name=None,
        status=None,
        created_after=None
    ):
        """Get all the refreshes that match the specified criteria. If no
        criterion is specified, it returns all refreshes.

        Args:
            namespace_name (Optional[str]): namespace the topics belong to.
            status (Optional[RefreshStatus]): The status associated with the refresh.
            created_after (Optional[int]): Epoch timestamp the refreshes should
                be created after. The refreshes created at the same timestamp
                are also included.

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.refresh.Refresh]):
                list of refreshes that match the given criteria.
        """
        response = self._call_api(
            api=self._client.refreshes.get_refreshes_by_criteria,
            params={
                'namespace': namespace_name,
                'status': status.value if status is not None else None,
                'created_after': created_after
            }
        )
        return [self._get_refresh_result_from_response(resp) for resp in response]

    def create_refresh(
        self,
        source_id,
        offset,
        batch_size,
        priority,
        filter_condition=None,
        avg_rows_per_second_cap=None
    ):
        """Register a refresh and returns the newly created refresh object.

        Args:
            source_id (int): The id of the source of the refresh.
            offset (int): The last known offset that has been refreshed.
            batch_size (int): The number of rows to be refreshed per batch.
            priority (Priority): The priority of the refresh
            filter_condition (Optional[str]): The filter condition associated with
             the refresh.
            avg_rows_per_second_cap (Optional[int]): Throughput throttle of the refresh.

        Returns:
            (data_pipeline.schematizer_clientlib.models.refresh.Refresh):
                The newly created Refresh.
        """
        request_body = {
            'offset': offset,
            'batch_size': batch_size,
            'priority': priority.name
        }
        if filter_condition:
            request_body['filter_condition'] = filter_condition
        if avg_rows_per_second_cap is not None:
            request_body['avg_rows_per_second_cap'] = avg_rows_per_second_cap
        response = self._call_api(
            api=self._client.sources.create_refresh,
            params={'source_id': source_id},
            request_body=request_body
        )
        return self._get_refresh_result_from_response(response)

    def update_refresh(self, refresh_id, status, offset):
        """Update the status of the refresh with specified refresh_id.

        Args:
            refresh_id (int): The id of the refresh being updated.
            status (RefreshStatus): New status of the refresh.
            offset (int): Last known offset that has been refreshed.

        Returns:
            (data_pipeline.schematizer_clientlib.models.refresh.Refresh):
                The updated Refresh.
        """
        request_body = {
            'status': status.value,
            'offset': offset
        }
        response = self._call_api(
            api=self._client.refreshes.update_refresh,
            params={'refresh_id': refresh_id},
            request_body=request_body
        )
        return self._get_refresh_result_from_response(response)

    def get_refreshes_by_namespace(self, namespace_name):
        """"Get the list of refreshes in the specified namespace.

        Args:
            namespace_name (str): namespace name to look up.

        Returns:
            (List[data_pipeline.schematizer_clientlib.models.refresh.Refresh]):
                The list of refreshes in the given namespace.
        """
        response = self._call_api(
            api=self._client.namespaces.list_refreshes_by_namespace,
            params={'namespace': namespace_name}
        )
        return [self._get_refresh_result_from_response(resp) for resp in response]

    def get_refresh_by_id(self, refresh_id):
        """Get the refresh associated with specified refresh_id.

        Args:
            refresh_id (int): The id of the refresh.

        Returns:
            (data_pipeline.schematizer_clientlib.models.refresh.Refresh):
                The refresh with the given refresh_id.
        """
        response = self._call_api(
            api=self._client.refreshes.get_refresh_by_id,
            params={'refresh_id': str(refresh_id)}
        )
        return self._get_refresh_result_from_response(response)

    def create_consumer_group(self, group_name, data_target_id):
        """ Create and return newly created consumer group.

        Args:
            group_name (str): The name of the new consumer group.
            data_target_id (int): The id of the data target which the new consumer
                group associates to.

        Returns:
            (data_pipeline.schematizer_clientlib.models.consumer_group.ConsumerGroup):
                The newly created consumer group.
        """
        response = self._call_api(
            api=self._client.data_targets.create_consumer_group,
            params={'data_target_id': data_target_id},
            request_body={'group_name': group_name}
        )
        _consumer_group = _ConsumerGroup.from_response(response)
        self._set_cache_by_consumer_group(_consumer_group)
        return _consumer_group.to_result()

    def get_consumer_group_by_id(self, consumer_group_id):
        """Get the consumer group of specified id.

        Args:
            consumer_group_id (int): The id of requested consumer group.

        Returns:
            (data_pipeline.consumer_grouptizer_clientlib.models.consumer_group.ConsumerGroup):
                The requested consumer group.
        """
        return self._get_consumer_group_by_id(consumer_group_id).to_result()

    def _get_consumer_group_by_id(self, consumer_group_id):
        _consumer_group = self._get_cached_consumer_group(consumer_group_id)
        if _consumer_group:
            return _consumer_group

        response = self._call_api(
            api=self._client.consumer_groups.get_consumer_group_by_id,
            params={'consumer_group_id': consumer_group_id}
        )
        _consumer_group = _ConsumerGroup.from_response(response)
        self._set_cache_by_consumer_group(_consumer_group)
        return _consumer_group

    def create_consumer_group_data_source(
        self,
        consumer_group_id,
        data_source_type,
        data_source_id
    ):
        """ Create and return newly created mapping between specified data
        source and consumer group.

        Args:
            consumer_group_id (int): The id of the consumer group.
            data_source_type
            (data_pipeline.schematizer_client.models.data_source_type_enum.DataSourceTypeEnum):
                Type of the data source.
            data_source_id (int): The id of the data source, which could be a
            namespace id or source id.

        Returns:
            (data_pipeline.schematizer_clientlib.models.consumer_group_data_source
            .ConsumerGroupDataSource):
            The newly created mapping between specified consumer group and
            data source.
        """
        response = self._call_api(
            api=self._client.consumer_groups.create_consumer_group_data_source,
            params={'consumer_group_id': consumer_group_id},
            request_body={
                'data_source_type': data_source_type.name,
                'data_source_id': data_source_id
            }
        )
        _consumer_group_data_src = _ConsumerGroupDataSource.from_response(response)
        return _consumer_group_data_src.to_result()

    def is_avro_schema_compatible(
        self,
        avro_schema_str,
        source_name,
        namespace_name
    ):
        """Determines if given avro_schema is backward and forward compatible with all
        enabled schemas of given source (contained in given namespace).

        Note: Compatibility means the input schema should be able to deserialize data serialized
            by existing schemas within the same topic and vice versa.

        Args:
            avro_schema_str (str): json string representing avro_schema to check compatiblity on.
            source_name (str): name of the source that contains the schemas to check compatibiliy
                against.
            namespace_name (str): name of namespace containing the given source

        Returns:
            (boolean): If the given schema is compatible with all enabled schemas of the source."""
        response = self._call_api(
            api=self._client.compatibility.is_avro_schema_compatible,
            request_body={
                'source': source_name,
                'namespace': namespace_name,
                'schema': avro_schema_str
            }
        )
        return response

    def filter_topics_by_pkeys(self, topics):
        """ Create and return a new list of topic names built from topics,
        filtered by whether a topic's most recent schema has a primary_key

        Args:
            topics (list[str]): List of topic names

        Returns:
            Newly created list of topic names filtered by if the corresponding
            topics have primary_keys in their most recent schemas
        """
        pkey_topics = []
        for topic in topics:
            try:
                schema = self.get_latest_schema_by_topic_name(topic)
                if schema.primary_keys:
                    pkey_topics.append(topic)
            except HTTPNotFound:
                # List of topics may include topics not in schematizer
                pass
        return pkey_topics

    def get_schema_migration(self, new_schema, target_schema_type, old_schema=None):
        """ Get a list of of SQL statements needed to migrate to a desired avro schema
        from an old avro schema. In the absence of an old avro schema, the migration
        generates a plan to just create the new schema.

        Args:
            new_schema (dict): The avro schema to which we want to migrate
            target_schema_type
            (data_pipeline.schematizer_client.models.target_schema_type_enum.TargetSchemaTypeEnum):
                Type of the target schema.
            old_schema (Optional[dict]): The avro schema from which we want to migrate

        Returns:
            (data_pipeline.schematizer_clientlib.models.schema_migration
            .SchemaMigration):
            An object containing the pushplan required to migrate to the desired schema
        """
        request_body = {
            'new_schema': simplejson.dumps(new_schema),
            'target_schema_type': target_schema_type.name,
        }
        if old_schema:
            request_body['old_schema'] = simplejson.dumps(old_schema)

        response = self._call_api(
            api=self._client.schema_migrations.get_schema_migration,
            request_body=request_body
        )
        return response

    def _call_api(self, api, params=None, request_body=None):
        request_params = params or {}
        if request_body:
            request_params['body'] = request_body
        request = api(**request_params)
        retry_policy = RetryPolicy(
            ExpBackoffPolicy(with_jitter=True),
            max_retry_count=get_config().schematizer_client_max_connection_retry
        )
        response = retry_on_exception(
            retry_policy=retry_policy,
            retry_exceptions=RequestException,
            func_to_retry=self._get_api_result,
            request=request
        )
        return response

    def _get_api_result(self, request):
        return request.result()

    def _get_cached_schema(self, schema_id):
        _schema = self._cache.get_value(_AvroSchema, schema_id)
        if _schema:
            _schema.topic = self._get_topic_by_name(_schema.topic.name)
        return _schema

    def _set_cache_by_schema(self, new_schema):
        self._cache.set_value(new_schema.schema_id, new_schema)
        self._set_cache_by_topic(new_schema.topic)

    def _get_cached_topic(self, topic_name):
        _topic = self._cache.get_value(_Topic, topic_name)
        if _topic:
            _topic.source = self._get_source_by_id(_topic.source.source_id)
        return _topic

    def _set_cache_by_topic(self, new_topic):
        self._cache.set_value(new_topic.name, new_topic)
        self._set_cache_by_source(new_topic.source)

    def _set_cache_by_source(self, new_source):
        self._cache.set_value(new_source.source_id, new_source)

    def _set_cache_by_data_target(self, new_data_target):
        self._cache.set_value(new_data_target.data_target_id, new_data_target)

    def _set_cache_by_data_target_name(self, new_data_target):
        self._cache.set_value(new_data_target.name, new_data_target)

    def _get_cached_consumer_group(self, consumer_group_id):
        _consumer_group = self._cache.get_value(_ConsumerGroup, consumer_group_id)
        if _consumer_group:
            _consumer_group.data_target = self._get_data_target_by_id(
                _consumer_group.data_target.data_target_id
            )
        return _consumer_group

    def _set_cache_by_consumer_group(self, new_consumer_group):
        self._cache.set_value(
            new_consumer_group.consumer_group_id,
            new_consumer_group
        )
        self._set_cache_by_data_target(new_consumer_group.data_target)

    def _get_refresh_result_from_response(self, response):
        _refresh = _Refresh.from_response(response)
        return _refresh.to_result()


def get_schematizer():
    return SchematizerClient()



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import namedtuple

from data_pipeline.schematizer_clientlib.models.model_base import BaseModel
from data_pipeline.schematizer_clientlib.models.source import _Source


"""
Represent the data of a topic.  A topic represents a kafka topic which a
producer publishes messages to (or a consumer consumes messages from).

Args:
    topic_id (int): The id of the topic.
    name (str): The name of the topic.
    source (data_pipeline.schematizer_clientlib.models.source.Source):
        The source of the topic.
    contains_pii (bool): Whether the messages in this topic contain PII data.
    cluster_type (str): Kafka cluster type to connect. Ex datapipe, scribe, etc.
    primary_keys ([str]): List of keys that are the primary keys of the topics schemas
    created_at (str): The timestamp when the topic is created in ISO-8601
        format.
    updated_at (str): The timestamp when the topic is last updated in ISO-8601
        format.
"""
Topic = namedtuple(
    'Topic',
    ['topic_id', 'name', 'source', 'contains_pii', 'cluster_type',
     'primary_keys', 'created_at', 'updated_at']
)


class _Topic(BaseModel):

    def __init__(
        self,
        topic_id,
        name,
        source,
        contains_pii,
        cluster_type,
        primary_keys,
        created_at,
        updated_at
    ):
        self.topic_id = topic_id
        self.name = name
        self.source = source
        self.contains_pii = contains_pii
        self.cluster_type = cluster_type
        self.primary_keys = primary_keys
        self.created_at = created_at
        self.updated_at = updated_at

    @classmethod
    def from_response(cls, response):
        return cls(
            topic_id=response.topic_id,
            name=response.name,
            source=_Source.from_response(response.source),
            contains_pii=response.contains_pii,
            cluster_type=response.cluster_type,
            primary_keys=response.primary_keys,
            created_at=response.created_at,
            updated_at=response.updated_at
        )

    def to_cache_value(self):
        return {
            'topic_id': self.topic_id,
            'name': self.name,
            'source_id': self.source.source_id,
            'contains_pii': self.contains_pii,
            'cluster_type': self.cluster_type,
            'primary_keys': self.primary_keys,
            'created_at': self.created_at,
            'updated_at': self.updated_at
        }

    @classmethod
    def from_cache_value(cls, cache_value):
        id_only_source = _Source(
            source_id=cache_value['source_id'],
            name=None,
            owner_email=None,
            namespace=None,
            category=None
        )
        return cls(
            topic_id=cache_value['topic_id'],
            name=cache_value['name'],
            source=id_only_source,
            contains_pii=cache_value['contains_pii'],
            cluster_type=cache_value['cluster_type'],
            primary_keys=cache_value['primary_keys'],
            created_at=cache_value['created_at'],
            updated_at=cache_value['updated_at']
        )

    def to_result(self):
        return Topic(
            topic_id=self.topic_id,
            name=self.name,
            source=self.source.to_result(),
            contains_pii=self.contains_pii,
            cluster_type=self.cluster_type,
            primary_keys=self.primary_keys,
            created_at=self.created_at,
            updated_at=self.updated_at
        )



# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import simplejson

from data_pipeline.tools.introspector.base_command import IntrospectorCommand
from data_pipeline.tools.introspector.models import IntrospectorSchema
from data_pipeline.tools.introspector.models import IntrospectorTopic


class TopicInfoCommand(IntrospectorCommand):
    @classmethod
    def add_parser(cls, subparsers):
        info_command_parser = subparsers.add_parser(
            "topic",
            description="Get information on a specific data pipeline topic.",
            add_help=False
        )

        cls.add_base_arguments(info_command_parser)

        info_command_parser.add_argument(
            "topic_name",
            type=str,
            help="Name of topic to retrieve information on."
        )

        info_command_parser.set_defaults(
            command=lambda args: cls("data_pipeline_instropsector_info_topic").run(
                args,
                info_command_parser
            )
        )

    def info_topic(self, name):
        topic = self.schematizer.get_topic_by_name(name)
        topic = IntrospectorTopic(
            topic,
            kafka_topics=self._kafka_topics,
            topics_to_range_map=self._topics_with_messages_to_range_map
        ).to_ordered_dict()
        topic['schemas'] = self.list_schemas(name)
        return topic

    def list_schemas(
        self,
        topic_name
    ):
        schemas = self.schematizer.get_schemas_by_topic(topic_name)
        schemas = [IntrospectorSchema(schema).to_ordered_dict() for schema in schemas]
        schemas.sort(key=lambda schema: schema['created_at'], reverse=True)
        return schemas

    def process_args(self, args, parser):
        super(TopicInfoCommand, self).process_args(args, parser)
        self.topic_name = args.topic_name

    def run(self, args, parser):
        self.process_args(args, parser)
        print simplejson.dumps(
            self.info_topic(self.topic_name)
        )
